import numpy as np
from cpymad.madx import Madx

import xtrack as xt

import matplotlib.pyplot as plt

from cpymad.madx import Madx

mad = Madx()

# Load mad model and apply element shifts
mad.input('''
call, file = '../../test_data/psb_chicane/psb.seq';
call, file = '../../test_data/psb_chicane/psb_fb_lhc.str';

beam, particle=PROTON, pc=0.5708301551893517;
use, sequence=psb1;

select,flag=error,clear;
select,flag=error,pattern=bi1.bsw1l1.1*;
ealign, dx=-0.0057;

select,flag=error,clear;
select,flag=error,pattern=bi1.bsw1l1.2*;
select,flag=error,pattern=bi1.bsw1l1.3*;
select,flag=error,pattern=bi1.bsw1l1.4*;
ealign, dx=-0.0442;

twiss;
''')

line = xt.Line.from_madx_sequence(
    sequence=mad.sequence.psb1,
    allow_thick=True,
    enable_align_errors=True,
    deferred_expressions=True,
)
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV,
                            gamma0=mad.sequence.psb1.beam.gamma)
#!end-doc-part

line.configure_bend_model(core='full')
line.to_json('psb_00_from_mad.json')



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import xtrack as xt

# Load a line and build tracker
line = xt.Line.from_json(
    '../../test_data/hllhc15_noerrors_nobb/line_and_particle.json')
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1, energy0=7e12)
line.build_tracker()

# Periodic twiss
tw_p = line.twiss()

# Twiss over a range with user-defined initial conditions (at start)
tw1 = line.twiss(start='ip5', end='mb.c24r5.b1',
                betx=0.15, bety=0.15, py=1e-6)


# Twiss over a range with user-defined initial conditions at end
tw2 = line.twiss(start='ip5', end='mb.c24r5.b1', init_at=xt.END,
                alfx=3.50482, betx=131.189, alfy=-0.677173, bety=40.7318,
                dx=1.22515, dpx=-0.0169647)

# Twiss over a range with user-defined initial conditions at arbitrary location
tw3 = line.twiss(start='ip5', end='mb.c24r5.b1', init_at='mb.c14r5.b1',
                 alfx=-0.437695, betx=31.8512, alfy=-6.73282, bety=450.454,
                 dx=1.22606, dpx=-0.0169647)

# Initial conditions can also be taken from an existing twiss table
tw4 = line.twiss(start='ip5', end='mb.c24r5.b1', init_at='mb.c14r5.b1',
                 init=tw_p)

# More explicitly, a `TwissInit` object can be extracted from the twiss table
# and used as initial conditions
tw_init = tw_p.get_twiss_init('mb.c14r5.b1',)
tw5 = line.twiss(start='ip5', end='mb.c24r5.b1', init=tw_init)


#!end-doc-part

# Plot

# Choose the twiss to plot
tw = tw5

import matplotlib.pyplot as plt
plt.close('all')

fig1 = plt.figure(1, figsize=(6.4, 4.8*1.5))
spbet = plt.subplot(3,1,1)
spco = plt.subplot(3,1,2, sharex=spbet)
spdisp = plt.subplot(3,1,3, sharex=spbet)

spbet.plot(tw.s, tw.betx)
spbet.plot(tw.s, tw.bety)
spbet.set_ylabel(r'$\beta_{x,y}$ [m]')

spco.plot(tw.s, tw.x)
spco.plot(tw.s, tw.y)
spco.set_ylabel(r'(Closed orbit)$_{x,y}$ [m]')

spdisp.plot(tw.s, tw.dx)
spdisp.plot(tw.s, tw.dy)
spdisp.set_ylabel(r'$D_{x,y}$ [m]')
spdisp.set_xlabel('s [m]')

for nn in ['ip5', 'mb.c14r5.b1', 'mb.c24r5.b1']:
    for ax in [spbet, spco, spdisp]:
        ax.axvline(tw_p['s', nn], color='k', ls='--', alpha=.5)
    spbet.text(tw_p['s', nn], 22000, nn, rotation=90,
        horizontalalignment='right', verticalalignment='top', alpha=.5)

fig1.subplots_adjust(left=.15, right=.92, hspace=.27)
plt.show()


# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import pathlib

import numpy as np
from cpymad.madx import Madx

import xobjects as xo
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts

test_data_folder = pathlib.Path(
        __file__).parent.joinpath('../test_data').absolute()

@for_all_test_contexts
def test_line_with_second_order_maps(test_context):

    line = xt.Line.from_json(test_data_folder /
                             'hllhc15_thick/lhc_thick_with_knobs.json')
    line.build_tracker(_context=test_context)
    line.vars['vrf400'] = 16
    line.vars['lagrf400.b1'] = 0.5

    line.vars['acbh22.l7b1'] = 15e-6
    line.vars['acbv21.l7b1'] = 10e-6

    ele_cut = ['ip1', 'ip2', 'ip3', 'ip4', 'ip5', 'ip6', 'ip7']

    line_maps = line.get_line_with_second_order_maps(split_at=ele_cut)
    line_maps.build_tracker(_context=test_context)

    line_maps.get_table().show()

    tw = line.twiss()
    tw_map = line_maps.twiss()

    xo.assert_allclose(tw_map.rows[ele_cut].s, tw.rows[ele_cut].s, rtol=0, atol=1e-12)

    xo.assert_allclose(tw_map.rows[ele_cut].x, tw.rows[ele_cut].x, rtol=0, atol=1e-12)
    xo.assert_allclose(tw_map.rows[ele_cut].px, tw.rows[ele_cut].px, rtol=0, atol=1e-12)
    xo.assert_allclose(tw_map.rows[ele_cut].y, tw.rows[ele_cut].y, rtol=0, atol=1e-12)
    xo.assert_allclose(tw_map.rows[ele_cut].py, tw.rows[ele_cut].py, rtol=0, atol=1e-12)
    xo.assert_allclose(tw_map.rows[ele_cut].zeta, tw.rows[ele_cut].zeta, rtol=0, atol=1e-10)
    xo.assert_allclose(tw_map.rows[ele_cut].delta, tw.rows[ele_cut].delta, rtol=0, atol=1e-12)

    xo.assert_allclose(tw_map.rows[ele_cut].betx, tw.rows[ele_cut].betx, rtol=1e-5, atol=0)
    xo.assert_allclose(tw_map.rows[ele_cut].alfx, tw.rows[ele_cut].alfx, rtol=1e-5, atol=1e-6)
    xo.assert_allclose(tw_map.rows[ele_cut].bety, tw.rows[ele_cut].bety, rtol=1e-5, atol=0)
    xo.assert_allclose(tw_map.rows[ele_cut].alfy, tw.rows[ele_cut].alfy, rtol=1e-5, atol=1e-6)

    xo.assert_allclose(np.mod(tw_map.qx, 1), np.mod(tw.qx, 1), rtol=0, atol=1e-7)
    xo.assert_allclose(np.mod(tw_map.qy, 1), np.mod(tw.qy, 1), rtol=0, atol=1e-7)
    xo.assert_allclose(tw_map.dqx, tw.dqx, rtol=0, atol=5e-2)
    xo.assert_allclose(tw_map.dqy, tw.dqy, rtol=0, atol=5e-2)
    xo.assert_allclose(tw_map.c_minus, tw.c_minus, rtol=0, atol=1e-5)
    xo.assert_allclose(tw_map.circumference, tw.circumference, rtol=0, atol=5e-9)


@for_all_test_contexts
def test_second_order_maps_against_madx(test_context):


    orbit_settings = {
        'acbh19.r3b1': 15e-6,
        'acbv20.r3b1': 10e-6,
        'acbv19.r3b2': 15e-6,
        'acbh20.r3b2': 10e-6,
    }

    # Generate Xsuite maps

    collider = xt.Environment.from_json(test_data_folder /
                            'hllhc15_thick/hllhc15_collider_thick.json')
    collider.vars.update(orbit_settings)
    collider['lhcb1'].twiss_default['method'] = '4d'
    collider['lhcb2'].twiss_default['method'] = '4d'
    collider.build_trackers(_context=test_context)

    map_b1 = xt.SecondOrderTaylorMap.from_line(
        line=collider.lhcb1, start='ip3', end='ip4')

    map_b4 = xt.SecondOrderTaylorMap.from_line(
        line=collider.lhcb2, start='ip4', end='ip3')
    map_b2_reflected = map_b4.scale_coordinates(scale_x=-1, scale_px=-1)

    # Generate MAD-X maps

    mad = Madx(stdout=False)
    mad.call(str(test_data_folder / "hllhc15_thick/lhc.seq"))
    mad.call(str(test_data_folder / "hllhc15_thick/hllhc_sequence.madx"))

    mad.input("""
        beam, sequence=lhcb1, particle=proton, pc=7000;
        beam, sequence=lhcb2, particle=proton, pc=7000, bv=-1;
    """)

    mad.call(str(test_data_folder / "hllhc15_thick/opt_round_150_1500.madx"))
    mad.globals.update(orbit_settings)

    mad.use(sequence="lhcb1")
    seq = mad.sequence.lhcb1
    mad.twiss()

    mad.input('''
    select, flag=sectormap, pattern='ip';
    twiss, sectormap, sectorpure, sectortable=secttab_b1;
    ''')
    sectmad_b1  = xt.Table(mad.table.secttab_b1)

    mad.input('''
        seqedit,sequence=lhcb2;flatten;reflect;flatten;endedit;
        use, sequence=lhcb2;
    ''')

    mad.input('''
    select, flag=sectormap, pattern='ip';
    twiss, sectormap, sectorpure, sectortable=secttab_b2;
    ''')
    sectmad_b2  = xt.Table(mad.table.secttab_b2)

    # Compare

    for line_name in ['lhcb1', 'lhcb2']:

        if line_name == 'lhcb1':
            xs_map = map_b1
            sectmad = sectmad_b1
            start = 'ip3'
            end = 'ip4'
            tw = collider.lhcb1.twiss()
        elif line_name == 'lhcb2':
            xs_map = map_b2_reflected
            sectmad = sectmad_b2
            start = 'ip4'
            end = 'ip3'
            tw = collider.lhcb2.twiss()
        else:
            raise ValueError(f'Unknown line_name: {line_name}')

        TT = xs_map.T
        RR = xs_map.R
        k = xs_map.k


        nemitt_x = 2.5e-6
        nemitt_y = 2.5e-6
        scale_in = [
            np.sqrt(tw['betx', start] * nemitt_x / tw['gamma0']),
            np.sqrt(tw['gamx', start] * nemitt_x / tw['gamma0']),
            np.sqrt(tw['bety', start] * nemitt_y / tw['gamma0']),
            np.sqrt(tw['gamy', start] * nemitt_y / tw['gamma0']),
            0.05,
            1e-3]

        scale_out = [
            np.sqrt(tw['betx', end] * nemitt_x / tw['gamma0']),
            np.sqrt(tw['gamx', end] * nemitt_x / tw['gamma0']),
            np.sqrt(tw['bety', end] * nemitt_y / tw['gamma0']),
            np.sqrt(tw['gamy', end] * nemitt_y / tw['gamma0']),
            0.05,
            1e-3]

        # Check k
        for ii in range(6):
            scaled_k = k[ii] / scale_out[ii]
            scaled_k_mad = sectmad[f'k{ii+1}', end] / scale_out[ii]
            # The following means that a the orbit kick is the same within 5e-5 sigmas
            xo.assert_allclose(scaled_k, scaled_k_mad, atol=5e-5, rtol=0)

        # Check R
        for ii in range(6):
            for jj in range(6):
                scaled_rr = RR[ii, jj] / scale_out[ii] * scale_in[jj]
                scaled_rr_mad = sectmad[f'r{ii+1}{jj+1}', end] * (
                                    scale_in[jj] / scale_out[ii])
                # The following means that a change of one sigma in jj results
                # in an error of less than 5e-4 sigmas on ii
                xo.assert_allclose(scaled_rr, scaled_rr_mad, atol=5e-4, rtol=0)

        # Check T
        for ii in range(6):
            for jj in range(6):
                for kk in range(6):
                    scaled_tt = (TT[ii, jj, kk]
                                / scale_out[ii] * scale_in[jj] * scale_in[kk])
                    scaled_tt_mad = (sectmad[f't{ii+1}{jj+1}{kk+1}', end]
                                    / scale_out[ii] * scale_in[jj] * scale_in[kk])
                    # The following means that a change of one sigma in jj, kk results
                    # in an error of less than 5e-4 sigmas on ii
                    xo.assert_allclose(scaled_tt, scaled_tt_mad, atol=5e-4, rtol=0)





import xtrack as xt
import xobjects as xo


from xtrack.beam_elements.magnets import MagnetDrift

md = MagnetDrift(length=1.0, k0=0.0, k1=0.0, h=0.0, drift_model=0)

p0 = xt.Particles(kinetic_energy0=50e6,
                 x=1e-3, y=2e-3, zeta=1e-2, px=10e-3, py=20e-3, delta=1e-2)

# Expanded drift
p_test = p0.copy()
p_ref = p0.copy()

eref = xt.Drift(length=1.0)
md.track(p_test)
eref.track(p_ref)

xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# Exact drift
md.drift_model = 1

p_test = p0.copy()
p_ref = p0.copy()

eref = xt.Solenoid(length=1.0) # Solenoid is exact drift when off
md.track(p_test)
eref.track(p_ref)

xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# Polar drift
md.drift_model = 2
md.h = 0.3

p_test = p0.copy()
p_ref = p0.copy()

eref = xt.Bend(length=1.0, h=0.3, k0=0)

md.track(p_test)
eref.track(p_ref)

# Combined function magnet
md.drift_model = 3
md.k0 = 0.2
md.k1 = 0.4

eref = xt.Bend(length=1.0, h=0.3, k0=0.2, k1=0.4)
eref.model = 'expanded'

p_test = p0.copy()
p_ref = p0.copy()

md.track(p_test)
eref.track(p_ref)

xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# Exact bend with curvature
md.drift_model = 4
md.k0 = 0.2
md.k1 = 0.

eref = xt.Bend(length=1.0, h=0.3, k0=0.2, k1=0.)
eref.model = 'bend-kick-bend'

p_test = p0.copy()
p_ref = p0.copy()

md.track(p_test)
eref.track(p_ref)

xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# Exact bend without curvature
md.drift_model = 5
md.k0 = 0.2
md.k1 = 0.
md.h = 0.

eref = xt.Bend(length=1.0, h=0., k0=0.2, k1=0.)
eref.model = 'bend-kick-bend'

p_test = p0.copy()
p_ref = p0.copy()

md.track(p_test)
eref.track(p_ref)

xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
from scipy.constants import c, e, m_p

from PyHEADTAIL.machines.synchrotron import Synchrotron


class EmptyObject(object):
    pass


class LHC(Synchrotron):

    def __init__(self, machine_configuration=None, optics_mode='smooth', **kwargs):


        pp = EmptyObject()
        pp.machine_configuration = machine_configuration
        pp.optics_mode = optics_mode

        pp.longitudinal_mode = 'non-linear'
        pp.alpha       = 3.225e-04
        pp.h_RF        = 35640
        pp.mass        = m_p
        pp.charge      = e
        pp.RF_at       = 'middle'

        if pp.machine_configuration == 'Injection':
            pp.p0          = 450e9 * e / c
            pp.p_increment = 0.
            pp.accQ_x      = 64.28
            pp.accQ_y      = 59.31
            pp.V_RF        = 6e6
            pp.dphi_RF     = 0.
        elif machine_configuration == '6.5_TeV_collision_tunes':
            pp.p0          = 6500e9 * e / c
            pp.p_increment = 0.
            pp.accQ_x      = 64.31
            pp.accQ_y      = 59.32
            pp.V_RF        = 12e6
            pp.dphi_RF     = 0.
        else:
            raise ValueError('machine_configuration not recognized!')

        if pp.optics_mode == 'smooth':
            if 's' in list(kwargs.keys()):
                raise ValueError('s vector cannot be provided if optics_mode = "smooth"')

            pp.n_segments = kwargs['n_segments']
            pp.circumference = 26658.8832

            pp.name = None

            pp.beta_x      = 92.7
            pp.D_x         = 0
            pp.beta_y      = 93.2
            pp.D_y         = 0

            pp.alpha_x     = None
            pp.alpha_y     = None

            pp.s = None

        elif pp.optics_mode == 'non-smooth':
            if 'n_segments' in list(kwargs.keys()):
                raise ValueError('n_segments cannot be provided if optics_mode = "non-smooth"')
            pp.n_segments = None
            pp.circumference = None

            pp.name        = kwargs['name']

            pp.beta_x      = kwargs['beta_x']
            pp.beta_y      = kwargs['beta_y']

            try:
                pp.D_x         = kwargs['D_x']
            except KeyError:
                pp.D_x         = 0 * np.array(kwargs['s'])
            try:
                pp.D_y         = kwargs['D_y']
            except KeyError:
                pp.D_y         = 0 * np.array(kwargs['s'])

            pp.alpha_x     = kwargs['alpha_x']
            pp.alpha_y     = kwargs['alpha_y']

            pp.s = kwargs['s']

        else:
            raise ValueError('optics_mode not recognized!')

        # detunings
        pp.Qp_x        = 0
        pp.Qp_y        = 0

        pp.app_x       = 0
        pp.app_y       = 0
        pp.app_xy      = 0

        pp.i_octupole_focusing = None
        pp.i_octupole_defocusing = None
        pp.octupole_knob = None

        for attr in list(kwargs.keys()):
            if kwargs[attr] is not None:
                if type(kwargs[attr]) is list or type(kwargs[attr]) is np.ndarray:
                    str2print = '[%s ...]'%repr(kwargs[attr][0])
                else:
                    str2print = repr(kwargs[attr])
                self.prints('Synchrotron init. From kwargs: %s = %s'
                            % (attr, str2print))
                if not hasattr(pp, attr):
                    raise NameError("I don't understand %s"%attr)

                setattr(pp, attr, kwargs[attr])



        if pp.i_octupole_focusing is not None or pp.i_octupole_defocusing is not None:
            if pp.octupole_knob is not None:
                raise ValueError('octupole_knobs and octupole currents cannot be used at the same time!')
            pp.app_x, pp.app_y, pp.app_xy = self._anharmonicities_from_octupole_current_settings(
                pp.i_octupole_focusing, pp.i_octupole_defocusing)
            self.i_octupole_focusing = pp.i_octupole_focusing
            self.i_octupole_defocusing = pp.i_octupole_defocusing

        if pp.octupole_knob is not None:
            if pp.i_octupole_focusing is not None or pp.i_octupole_defocusing is not None:
                raise ValueError('octupole_knobs and octupole currents cannot be used at the same time!')
            pp.i_octupole_focusing, pp.i_octupole_defocusing = self._octupole_currents_from_octupole_knobs(pp.octupole_knob, pp.p0)
            pp.app_x, pp.app_y, pp.app_xy = self._anharmonicities_from_octupole_current_settings(
                pp.i_octupole_focusing, pp.i_octupole_defocusing)
            self.i_octupole_focusing = pp.i_octupole_focusing
            self.i_octupole_defocusing = pp.i_octupole_defocusing


        super(LHC, self).__init__(optics_mode=pp.optics_mode, circumference=pp.circumference, n_segments=pp.n_segments,
             s=pp.s, name=pp.name,
             alpha_x=pp.alpha_x, beta_x=pp.beta_x, D_x=pp.D_x, alpha_y=pp.alpha_y, beta_y=pp.beta_y, D_y=pp.D_y,
             accQ_x=pp.accQ_x, accQ_y=pp.accQ_y, Qp_x=pp.Qp_x, Qp_y=pp.Qp_y, app_x=pp.app_x, app_y=pp.app_y, app_xy=pp.app_xy,
             alpha_mom_compaction=pp.alpha, longitudinal_mode=pp.longitudinal_mode,
             h_RF=np.atleast_1d(pp.h_RF), V_RF=np.atleast_1d(pp.V_RF), dphi_RF=np.atleast_1d(pp.dphi_RF),
             p0=pp.p0, p_increment=pp.p_increment,
             charge=pp.charge, mass=pp.mass, RF_at=pp.RF_at)


    def _anharmonicities_from_octupole_current_settings(self, i_octupole_focusing, i_octupole_defocusing):
            """Calculate the constants of proportionality app_x, app_y and
            app_xy (== app_yx) for the amplitude detuning introduced by the
            LHC octupole magnets (aka. LHC Landau octupoles) from the
            electric currents i_octupole_focusing [A] and i_octupole_defocusing [A] flowing
            through the magnets. The maximum current is given by
            i_max = +/- 550 [A]. The values app_x, app_y, app_xy obtained
            from the formulae are proportional to the strength of detuning
            for one complete turn around the accelerator, i.e. one-turn
            values.
            The calculation is based on formulae (3.6) taken from 'The LHC
            transverse coupled-bunch instability' by N. Mounet, EPFL PhD
            Thesis, 2012. Values (hard-coded numbers below) are valid for
            LHC Landau octupoles before LS1. Beta functions in x and y are
            correctly taken into account. Note that here, the values of
            app_x, app_y and app_xy are not normalized to the reference
            momentum p0. This is done only during the calculation of the
            detuning in the corresponding detune method of the
            AmplitudeDetuningSegment.
            More detailed explanations and references on how the formulae
            were obtained are given in the PhD thesis (pg. 85ff) cited
            above.
            """
            i_max = 550.  # [A]
            E_max = 7000. # [GeV]

            app_x  = E_max * (267065. * i_octupole_focusing / i_max -
                7856. * i_octupole_defocusing / i_max)
            app_y  = E_max * (9789. * i_octupole_focusing / i_max -
                277203. * i_octupole_defocusing / i_max)
            app_xy = E_max * (-102261. * i_octupole_focusing / i_max +
                93331. * i_octupole_defocusing / i_max)

            # Convert to SI units.
            convert_to_SI = e / (1.e-9 * c)
            app_x *= convert_to_SI
            app_y *= convert_to_SI
            app_xy *= convert_to_SI

            return app_x, app_y, app_xy

    def _octupole_currents_from_octupole_knobs(self, octupole_knob, p0):
        i_octupole_focusing = 19.557 * octupole_knob / (-1.5) * p0 / 2.4049285931335872e-16
        i_octupole_defocusing = - i_octupole_focusing
        return i_octupole_focusing, i_octupole_defocusing



import numpy as np
import xtrack as xt

pi = np.pi
lbend = 3

# Create an environment
env = xt.Environment()

# Build a simple ring
line = env.new_line(components=[
    env.new('mqf.1', xt.Quadrupole, length=0.3, k1=0.1),
    env.new('d1.1',  xt.Drift, length=1),
    env.new('mb1.1', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d2.1',  xt.Drift, length=1),

    env.new('mqd.1', xt.Quadrupole, length=0.3, k1=-0.7),
    env.new('d3.1',  xt.Drift, length=1),
    env.new('mb2.1', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d4.1',  xt.Drift, length=1),

    env.new('mqf.2', xt.Quadrupole, length=0.3, k1=0.1),
    env.new('d1.2',  xt.Drift, length=1),
    env.new('mb1.2', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d2.2',  xt.Drift, length=1),

    env.new('mqd.2', xt.Quadrupole, length=0.3, k1=-0.7),
    env.new('d3.2',  xt.Drift, length=1),
    env.new('mb2.2', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d4.2',  xt.Drift, length=1),
])

# Build the ring
line.particle_ref = xt.Particles(p0c=1.2e9, mass0=xt.PROTON_MASS_EV)
line.build_tracker()

# Inspect the line
tab = line.get_table()
tab.show()
# prints:
#
# name          s element_type isthick isreplica parent_name iscollective
# mqf.1         0 Quadrupole      True     False        None        False
# d1.1        0.3 Drift           True     False        None        False
# mb1.1       1.3 Bend            True     False        None        False
# d2.1        4.3 Drift           True     False        None        False
# mqd.1       5.3 Quadrupole      True     False        None        False
# d3.1        5.6 Drift           True     False        None        False
# mb2.1       6.6 Bend            True     False        None        False
# d4.1        9.6 Drift           True     False        None        False
# mqf.2      10.6 Quadrupole      True     False        None        False
# d1.2       10.9 Drift           True     False        None        False
# mb1.2      11.9 Bend            True     False        None        False
# d2.2       14.9 Drift           True     False        None        False
# mqd.2      15.9 Quadrupole      True     False        None        False
# d3.2       16.2 Drift           True     False        None        False
# mb2.2      17.2 Bend            True     False        None        False
# d4.2       20.2 Drift           True     False        None        False
# _end_point 21.2                False     False        None        False

# Define a sextupole
my_sext = xt.Sextupole(length=0.1, k2=0.1)
# Insert copies of the defined sextupole downstream of the quadrupoles
line.discard_tracker() # needed to modify the line structure
line.insert('msf.1', my_sext.copy(), anchor='start', at='mqf.1@end')
line.insert('msd.1', my_sext.copy(), anchor='start', at='mqd.1@end')
line.insert('msf.2', my_sext.copy(), anchor='start', at='mqf.2@end')
line.insert('msd.2', my_sext.copy(), anchor='start', at='mqd.2@end')

# Define a rectangular aperture
my_aper = xt.LimitRect(min_x=-0.02, max_x=0.02, min_y=-0.01, max_y=0.01)
# Insert the aperture upstream of the first bending magnet
line.insert('aper', my_aper, at='mb1.1@start')

line.get_table().show()
# prints:
#
# name          s element_type isthick isreplica parent_name iscollective
# mqf.1         0 Quadrupole      True     False        None        False
# d1.1..0     0.3 DriftSlice      True     False        d1.1        False
# msf.1       0.4 Sextupole       True     False        None        False
# d1.1..2     0.5 DriftSlice      True     False        d1.1        False
# aper        1.3 LimitRect      False     False        None        False
# mb1.1       1.3 Bend            True     False        None        False
# d2.1        4.3 Drift           True     False        None        False
# mqd.1       5.3 Quadrupole      True     False        None        False
# d3.1..0     5.6 DriftSlice      True     False        d3.1        False
# msd.1       5.7 Sextupole       True     False        None        False
# d3.1..2     5.8 DriftSlice      True     False        d3.1        False
# mb2.1       6.6 Bend            True     False        None        False
# d4.1        9.6 Drift           True     False        None        False
# mqf.2      10.6 Quadrupole      True     False        None        False
# d1.2..0    10.9 DriftSlice      True     False        d1.2        False
# msf.2        11 Sextupole       True     False        None        False
# d1.2..2    11.1 DriftSlice      True     False        d1.2        False
# mb1.2      11.9 Bend            True     False        None        False
# d2.2       14.9 Drift           True     False        None        False
# mqd.2      15.9 Quadrupole      True     False        None        False
# d3.2..0    16.2 DriftSlice      True     False        d3.2        False
# msd.2      16.3 Sextupole       True     False        None        False
# d3.2..2    16.4 DriftSlice      True     False        d3.2        False
# mb2.2      17.2 Bend            True     False        None        False
# d4.2       20.2 Drift           True     False        None        False
# _end_point 21.2                False     False        None        False



import numpy as np
from pathlib import Path

from scipy.constants import c as clight
from scipy.constants import e as qe
from scipy.constants import epsilon_0 as eps0

import xobjects as xo
import xtrack as xt
from xtrack._temp.boris_and_solenoid_map.solenoid_field import SolenoidField


ctx = xo.ContextCpu()

boris_knl_description = xo.Kernel(
    c_name='boris_step',
    args=[
        xo.Arg(xo.Int64,   name='N_sub_steps'),
        xo.Arg(xo.Float64, name='Dtt'),
        xo.Arg(xo.Float64, name='B_field', pointer=True),
        xo.Arg(xo.Float64, name='B_skew', pointer=True),
        xo.Arg(xo.Float64, name='xn1', pointer=True),
        xo.Arg(xo.Float64, name='yn1', pointer=True),
        xo.Arg(xo.Float64, name='zn1', pointer=True),
        xo.Arg(xo.Float64, name='vxn1', pointer=True),
        xo.Arg(xo.Float64, name='vyn1', pointer=True),
        xo.Arg(xo.Float64, name='vzn1', pointer=True),
        xo.Arg(xo.Float64, name='Ex_n', pointer=True),
        xo.Arg(xo.Float64, name='Ey_n', pointer=True),
        xo.Arg(xo.Float64, name='Bx_n_custom', pointer=True),
        xo.Arg(xo.Float64, name='By_n_custom', pointer=True),
        xo.Arg(xo.Float64, name='Bz_n_custom', pointer=True),
        xo.Arg(xo.Int64,   name='custom_B'),
        xo.Arg(xo.Int64,   name='N_mp'),
        xo.Arg(xo.Int64,   name='N_multipoles'),
        xo.Arg(xo.Float64, name='charge'),
        xo.Arg(xo.Float64, name='mass', pointer=True),
    ],
)

ctx.add_kernels(
    kernels={'boris': boris_knl_description},
    sources=[xt._pkg_root / '_temp/boris_and_solenoid_map/boris.h'],
)

# tag = 'case0'
# orbit_lims = -5, 5
# p0 = xt.Particles(mass0=xt.ELECTRON_MASS_EV, q0=1,
#                  energy0=45.6e9,
#                  x=1e-3, px=0,
#                  y=1e-3, py=0,
#                  delta=0)

# tag = 'case1'
# orbit_lims = -100, 100
# p0 = xt.Particles(mass0=xt.ELECTRON_MASS_EV, q0=1,
#                  energy0=45.6e9,
#                  x=-170e-3, px=15e-3,
#                  y=10e-3, py=-5e-3,
#                  delta=0)

# tag = 'case2'
# orbit_lims = -100, 100
# delta = -0.1
# p0 = xt.Particles(mass0=xt.ELECTRON_MASS_EV, q0=1,
#                  energy0=45.6e9,
#                  x=-170e-3, px=15e-3 * (1+delta),
#                  y=10e-3, py=-5e-3 * (1+delta),
#                  delta=delta)

# tag = 'case3'
# orbit_lims = -100, 100
# delta = -0.99
# p0 = xt.Particles(mass0=xt.ELECTRON_MASS_EV, q0=1,
#                  energy0=45.6e9,
#                  x=-170e-3, px=15e-3 * (1+delta),
#                  y=10e-3, py=-5e-3 * (1+delta),
#                  delta=delta)

# tag = 'case4'
# orbit_lims = -100, 100
# delta = -0.999
# p0 = xt.Particles(mass0=xt.ELECTRON_MASS_EV, q0=1,
#                  energy0=45.6e9,
#                  x=-170e-3, px=15e-3 * (1+delta),
#                  y=10e-3, py=-5e-3 * (1+delta),
#                  delta=delta)

tag = 'case5'
orbit_lims = -100, 100
p0 = xt.Particles(mass0=xt.ELECTRON_MASS_EV, q0=1,
                 energy0=45.6e9,
                 x=-150e-3, px=15e-3,
                 y=30e-3, py=-5e-3,
                 delta=0)

p = p0.copy()

z_sol_center = 10
sf = SolenoidField(L=4, a=0.3, B0=1., z0=z_sol_center)

dt = 1e-10
n_steps = 1500

x_log = []
y_log = []
z_log = []
px_log = []
py_log = []
pp_log = []
pow_log = []
pow_x_log = []
pow_y_log = []
Bx_log = []
By_log = []
Bz_log = []

for ii in range(n_steps):

    x = p.x.copy()
    y = p.y.copy()
    z = p.s.copy()

    Bx, By, Bz = sf.get_field(x, y, z)

    gamma = p.energy / p.mass0
    mass0_kg = p.mass0 * qe / clight**2
    charge0_coulomb = p.q0 * qe

    p0c_J = p.p0c * qe

    Pxc_J = p.px * p0c_J
    Pyc_J = p.py * p0c_J
    Pzc_J = np.sqrt((p0c_J*(1 + p.delta))**2 - Pxc_J**2 - Pyc_J**2)

    vx = Pxc_J / clight / (gamma * mass0_kg) # m/s
    vy = Pyc_J / clight / (gamma * mass0_kg) # m/s
    vz = Pzc_J / clight / (gamma * mass0_kg) # m/s

    vx_before = vx.copy()
    vy_before = vy.copy()

    ctx.kernels.boris(
            N_sub_steps=1,
            Dtt=dt,
            B_field=np.array([0.]),
            B_skew=np.array([0.]),
            xn1=x,
            yn1=y,
            zn1=z,
            vxn1=vx,
            vyn1=vy,
            vzn1=vz,
            Ex_n=0 * x,
            Ey_n=0 * x,
            Bx_n_custom=Bx,
            By_n_custom=By,
            Bz_n_custom=Bz,
            custom_B=1,
            N_mp=len(x),
            N_multipoles=0,
            charge=charge0_coulomb,
            mass=mass0_kg * gamma,
    )

    p.x = x
    p.y = y
    p.s = z
    p.px = mass0_kg * gamma * vx * clight / p0c_J
    p.py = mass0_kg * gamma * vy * clight / p0c_J
    pz = mass0_kg * gamma * vz * clight / p0c_J
    pp = np.sqrt(p.px**2 + p.py**2 + pz**2)

    beta_x_before = vx_before / clight
    beta_y_before = vy_before / clight

    beta_x_after = vx / clight
    beta_y_after = vy / clight

    beta_x_dot = (beta_x_after - beta_x_before) / dt
    beta_y_dot = (beta_y_after - beta_y_before) / dt

    bet_dot_square = beta_x_dot**2 + beta_y_dot**2

    # From Hofmann, "The physics of synchrontron radiation" Eq 3.7 and below
    pow = 2 * qe**2 * bet_dot_square * gamma**4 / (12 * np.pi * eps0 * clight)

    x_log.append(p.x.copy())
    y_log.append(p.y.copy())
    z_log.append(p.s.copy())
    px_log.append(p.px.copy())
    py_log.append(p.py.copy())
    pp_log.append(pp)
    pow_log.append(pow)
    Bx_log.append(Bx)
    By_log.append(By)
    Bz_log.append(Bz)

x_log = np.array(x_log)
y_log = np.array(y_log)
z_log = np.array(z_log)
px_log = np.array(px_log)
py_log = np.array(py_log)
pp_log = np.array(pp_log)
Bx_log = np.array(Bx_log)
By_log = np.array(By_log)
Bz_log = np.array(Bz_log)

pow_log = np.array(pow_log)

dE_ds_boris_J = 0 * pow_log

dE_ds_boris_J[:-1]= pow_log[:-1] * dt / np.diff(z_log, axis=0)
dE_ds_boris_eV = dE_ds_boris_J / qe

z_axis = np.linspace(0, 30, 1001)
Bz_axis = sf.get_field(0 * z_axis, 0 * z_axis, z_axis)[2]

P0_J = p.p0c[0] * qe / clight
brho = P0_J / qe / p.q0

#ks = 0.5 * (Bz_axis[:-1] + Bz_axis[1:]) / brho
ks = Bz_axis[:-1] / brho

line = xt.Line(elements=[xt.Solenoid(length=z_axis[1]-z_axis[0], ks=ks[ii])
                            for ii in range(len(z_axis)-1)])
line.build_tracker()
line.configure_radiation(model='mean')

p_xt = p0.copy()
line.track(p_xt, turn_by_turn_monitor='ONE_TURN_EBE')
mon = line.record_last_track

p_xt = p0.copy()
line.configure_radiation(model=None)
line.track(p_xt, turn_by_turn_monitor='ONE_TURN_EBE')
mon_no_rad = line.record_last_track

Bz_mid = 0.5 * (Bz_axis[:-1] + Bz_axis[1:])
Bz_mon = 0 * Bz_axis
Bz_mon[1:] = Bz_mid

# Wolsky Eq. 3.114
Ax = -0.5 * Bz_mon * mon.y
Ay =  0.5 * Bz_mon * mon.x

# Wolsky Eq. 2.74
ax_ref = Ax * p0.q0 * qe / P0_J
ay_ref = Ay * p0.q0 * qe / P0_J

px_mech = mon.px - ax_ref
py_mech = mon.py - ay_ref
pz_mech = np.sqrt((1 + mon.delta)**2 - px_mech**2 - py_mech**2)

xp = px_mech / pz_mech
yp = py_mech / pz_mech

dx_ds = np.diff(mon.x, axis=1) / np.diff(mon.s, axis=1)
dy_ds = np.diff(mon.y, axis=1) / np.diff(mon.s, axis=1)

dE_ds = -np.diff(mon.ptau, axis=1)/np.diff(mon.s, axis=1) * p_xt.energy0[0]

emitted_dpx = -(np.diff(mon.px, axis=1) - np.diff(mon_no_rad.px, axis=1))
emitted_dpy = -(np.diff(mon.py, axis=1) - np.diff(mon_no_rad.py, axis=1))
emitted_dp = -(np.diff(mon.delta, axis=1) - np.diff(mon_no_rad.delta, axis=1))

z_check = sf.z0 + sf.L * np.linspace(-2, 2, 1001)

for i_part in range(z_log.shape[1]):

    this_s_boris = 0.5 * (z_log[:-1, i_part] + z_log[1:, i_part])
    dx_ds_boris = np.diff(x_log[:, i_part]) / np.diff(z_log[:, i_part])
    dy_ds_boris = np.diff(y_log[:, i_part]) / np.diff(z_log[:, i_part])

    s_xsuite = 0.5 * (mon.s[i_part, :-1] + mon.s[i_part, 1:])
    dx_ds_xsuite = np.diff(mon.x[i_part, :]) / np.diff(mon.s[i_part, :])
    dy_ds_xsuite = np.diff(mon.y[i_part, :]) / np.diff(mon.s[i_part, :])
    dE_ds_xsuite = dE_ds[i_part, :]

    dx_ds_xsuite_check = np.interp(z_check, s_xsuite, dx_ds_xsuite)
    dy_ds_xsuite_check = np.interp(z_check, s_xsuite, dy_ds_xsuite)
    dE_ds_xsuite_check = np.interp(z_check, s_xsuite, dE_ds_xsuite)

    dx_ds_boris_check = np.interp(z_check, this_s_boris, dx_ds_boris)
    dy_ds_boris_check = np.interp(z_check, this_s_boris, dy_ds_boris)
    dE_ds_boris_check = np.interp(z_check, z_log[:, i_part], dE_ds_boris_eV[:, i_part])

    this_emitted_dpx = emitted_dpx[i_part, :]
    this_emitted_dpy = emitted_dpy[i_part, :]
    this_dE_ds = dE_ds[i_part, :]
    this_dx_ds = dx_ds[i_part, :]
    this_dy_ds = dy_ds[i_part, :]

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
ax1 = plt.subplot(2, 1, 1)
plt.plot(z_log, x_log, label='Boris')
plt.plot(mon.s.T, mon.x.T, '--', label='xsuite')
plt.ylabel('x [m]')
plt.legend()

ax2 = plt.subplot(2, 1, 2, sharex=ax1)
plt.plot(z_axis, Bz_axis)
plt.ylabel(r'$B_{z}$ [T]')
plt.xlabel('z [m]')

plt.figure(2)
dx_ds_log = np.diff(x_log, axis=0) / np.diff(z_log, axis=0)
dy_ds_log = np.diff(y_log, axis=0) / np.diff(z_log, axis=0)
ax1 = plt.subplot(2, 1, 1)
plt.plot(0.5 * (z_log[:-1, :] + z_log[1:, :]), dx_ds_log, label='Boris')
plt.plot(mon.s.T, xp.T, label="x'", color='C1', linestyle='-')
plt.plot(mon.s[:, :-1].T, dx_ds.T, '--', color='C2',
         label=r"$\Delta x / \Delta s$")
plt.ylabel(r"$x'$")
plt.legend()

ax2 = plt.subplot(2, 1, 2, sharex=ax1)
plt.plot(0.5 * (z_log[:-1] + z_log[1:]), dy_ds_log, label='Boris')
plt.plot(mon.s.T, yp.T, label="y'", color='C1', linestyle='-')
plt.plot(mon.s[:, :-1].T, dy_ds.T, '--', color='C2',
         label=r"$\Delta y / \Delta s$")


# Compare ax and ay
plt.figure(3)
plt.plot(mon.s.T, ax_ref.T, label="ax_ref", color='C0', linestyle='-')
plt.plot(mon.s.T, ay_ref.T, label="ay_ref", color='C1', linestyle='-')
plt.plot(mon.s.T, mon.ax.T, label="ax", color='C2', linestyle='--')
plt.plot(mon.s.T, mon.ay.T, label="ay", color='C3', linestyle='--')

plt.figure(4)
plt.plot(mon.s[:, :-1].T, dE_ds.T * 1e-2 * 1e-3, '.-', label='dE/ds')
plt.plot(z_log, dE_ds_boris_eV * 1e-2 * 1e-3, 'x-', label='dE/ds Boris')

plt.figure(5)

ax1 = plt.subplot(2, 1, 1)
plt.plot(mon.s[:, :-1].T, emitted_dpx.T, '-', label='dpx')
plt.plot(mon.s[:, :-1].T, dE_ds.T * dx_ds.T*np.diff(mon.s, 1).T/p.p0c[0], '--')

ax2 = plt.subplot(2, 1, 2, sharex=ax1)
plt.plot(mon.s[:, :-1].T, emitted_dpy.T, '-', label='dpy')
plt.plot(mon.s[:, :-1].T, dE_ds.T * dy_ds.T*np.diff(mon.s, 1).T/p.p0c[0], '--')

i_part_plot = 0

plt.figure(100, figsize=(6.4*1.1, 4.8*1.7))


sp1 = plt.subplot(5, 1, 1)
# plt.plot(z_axis - z_sol_center, Bz_axis)
plt.plot(0.5 * (z_log[:-1, i_part] + z_log[1:, i_part]) - z_sol_center, Bz_log[:-1, i_part])
plt.ylabel(r'$B_{z}$ [T]')

sp2 = plt.subplot(5, 1, 2, sharex=sp1)
plt.plot(mon.s[i_part_plot, :] - z_sol_center, 1e3 * mon.x[i_part_plot, :])
plt.plot(mon.s[i_part_plot, :] - z_sol_center, 1e3 * mon.y[i_part_plot, :])
plt.axhline(0, color='grey', alpha=0.6, linestyle=':')
plt.ylim(*orbit_lims)
plt.ylabel('x, y [mm]')

px = mon.px[i_part_plot, :]
kin_px = mon.kin_px[i_part_plot, :]
kin_px_boris = px_log[:, i_part_plot]

py = mon.py[i_part_plot, :]
kin_py = mon.kin_py[i_part_plot, :]
kin_py_boris = py_log[:, i_part_plot]

sp3 = plt.subplot(5, 1, 3, sharex=sp1)
plt.plot(0.5 * (z_log[:-1, i_part] + z_log[1:, i_part]) - z_sol_center, Bx_log[:-1, i_part])
plt.plot(0.5 * (z_log[:-1, i_part] + z_log[1:, i_part]) - z_sol_center, By_log[:-1, i_part])
plt.ylabel(r'$B_{x,y}$ [T]')
plt.legend([r'$B_x$', r'$B_y$'], loc='upper left')

sp4 = plt.subplot(5, 1, 4, sharex=sp1)
plt.plot(mon.s[i_part_plot, :] - z_sol_center, 1e6 * (px - px[0]),
         label=r'$p_x$ (xsuite)')
plt.plot(mon.s[i_part_plot, :] - z_sol_center, 1e6 * (kin_px - kin_px[0]),
         label=r"$p_x^\text{kin}$ (xsuite)")
plt.plot(z_log[:, i_part_plot] - z_sol_center, 1e6 * (kin_px_boris - kin_px[0]),
         label=r"$p_x^\text{kin}$ (Boris)", linestyle='--')
plt.legend(fontsize='medium', loc='upper left')
plt.ylabel(r"$\Delta p_x$ [$10^{-6}$]")

sp5 = plt.subplot(5, 1, 5, sharex=sp1)
plt.plot(mon.s[i_part_plot, :] - z_sol_center, 1e6 * (py - py[0]),
         label=r'$p_y$ (xsuite)')
plt.plot(mon.s[i_part_plot, :] - z_sol_center, 1e6 * (kin_py - kin_py[0]),
         label=r"$p_y^\text{kin}$ (xsuite)")
plt.plot(z_log[:, i_part_plot] - z_sol_center, 1e6 * (kin_py_boris - kin_py[0]),
         label=r"$p_y^\text{kin}$ (Boris)", linestyle='--')
plt.legend(fontsize='medium', loc='upper left')
plt.ylabel(r"$\Delta p_y$ [$10^{-6}$]")

plt.xlim(-6, 6)
plt.xlabel('s [m]')

plt.subplots_adjust(top=.91, bottom=.06, hspace=.3, left=.14)
plt.suptitle(r'$\Delta s_\text{slices}$ = ' f'{np.diff(z_log, axis=0).mean():.2f}, '
             r'$E_0$ = ' f'{p.energy0[0]*1e-9:.1f} GeV, '
             r'$\delta$ = ' f'{p.delta[0]:.3f} '
             r'($E$ = ' f'{p.energy[0]*1e-9:.3f} GeV) \n'
             r"Initial $x'$ = " f"{mon.kin_xprime[i_part_plot, 0]*1e3:.2f} mrad, "
             r"initial $y'$ = " f"{mon.kin_yprime[i_part_plot, 0]*1e3:.2f} mrad")

plt.savefig(f'benchmark_{tag}.png', dpi=200)


plt.figure(101)
plt.plot(mon.s[i_part_plot, :-1] - z_sol_center, dE_ds[i_part_plot, :]* 1e-2,
            label='xsuite')
plt.plot(z_log[:, i_part_plot] - z_sol_center, dE_ds_boris_eV *1e-2,
         label="Boris", linestyle='--')
plt.ylabel(r"$dE/ds$ [eV/cm]")
plt.legend(fontsize='medium', loc='upper left')
plt.xlim(-6, 6)
plt.xlabel('s [m]')
plt.suptitle(r'$\Delta s_\text{slices}$ = ' f'{np.diff(z_log, axis=0).mean():.2f}, '
             r'$E_0$ = ' f'{p.energy0[0]*1e-9:.1f} GeV, '
             r'$\delta$ = ' f'{p.delta[0]:.3f} '
             r'($E$ = ' f'{p.energy[0]*1e-9:.3f} GeV) \n'
             r"Initial $x'$ = " f"{mon.kin_xprime[i_part_plot, 0]*1e3:.2f} mrad, "
             r"initial $y'$ = " f"{mon.kin_yprime[i_part_plot, 0]*1e3:.2f} mrad")

plt.show()


import numpy as np
import xtrack as xt

line = xt.Line.from_json(
    '../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')

tw = line.twiss(method='4d')

mng = line.to_madng(sequence_name='lhcb1')

mng.send('''
local damap in MAD
local lhc = MADX.lhcb1

-- list of octupolar RDTs
local rdts = {"f4000", "f3100", "f2020", "f1120"}

-- create phase-space damap at 4th order
local X0 = damap {nv=6, mo=4}

-- twiss with RDTs
local mtbl = twiss {sequence=lhc, X0=X0, trkrdt=rdts, info=2, saverdt=true}

-- send columns to Python
py:send({mtbl.s, mtbl.beta11, mtbl.f4000, mtbl.f3100, mtbl.f2020, mtbl.f1120})

''')


s, beta11, f4000, f3100, f2020, f1120 = mng.recv()

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)

plt.plot(s, np.abs(f4000), label='f4000')
plt.plot(s, np.abs(f2020), label='f2020')
plt.plot(s, np.abs(f1120), label='f1120')
plt.plot(s, np.abs(f3100), label='f3100')
plt.xlabel('s [m]')
plt.ylabel(r'|f_{jklm}|')
plt.legend()

plt.show()


import xtrack as xt

env = xt.load_madx_lattice('../../test_data/leir/leir.seq')
env.vars.load_madx('../../test_data/leir/leir_inj_nominal.str')

line = env.leir
line.configure_bend_model(edge='full', core='adaptive')
line.particle_ref = xt.Particles(kinetic_energy0=14e9)

from cpymad.madx import Madx
mad = Madx()

mad.call('../../test_data/leir/leir.seq')
mad.call('../../test_data/leir/leir_inj_nominal.str')
mad.beam()
mad.use(sequence='leir')

lref = xt.Line.from_madx_sequence(mad.sequence.leir)
lref.configure_bend_model(edge='full', core='adaptive')
lref.particle_ref = xt.Particles(kinetic_energy0=14e9)

tw_ref = lref.twiss4d(betx=1, bety=1)
tw = line.twiss4d(betx=1, bety=1)

mad.sequence.leir.expanded_elements['er.sol21'].cmdpar['ks'].expr # this is None ?!


import numpy as np
import xtrack as xt

# We build a simple ring
pi = np.pi
lbend = 3
lquad = 0.3
env = xt.Environment()
line = env.new_line(components=[
    env.new('mqf.1', xt.Quadrupole, length=lquad, k1=0.1),
    env.new('msf.1', xt.Sextupole, length=0.1, k2=0.02),
    env.new('d1.1',  xt.Drift, length=0.9),
    env.new('mb1.1', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d2.1',  xt.Drift, length=1),

    env.new('mqd.1', xt.Quadrupole, length=lquad, k1=-0.7),
    env.new('d3.1',  xt.Drift, length=1),
    env.new('mb2.1', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d4.1',  xt.Drift, length=1),

    env.new('mqf.2', xt.Quadrupole, length=lquad, k1=0.1),
    env.new('d1.2',  xt.Drift, length=1),
    env.new('mb1.2', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d2.2',  xt.Drift, length=1),

    env.new('mqd.2', xt.Quadrupole, length=lquad, k1=-0.7),
    env.new('d3.2',  xt.Drift, length=1),
    env.new('mb2.2', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d4.2',  xt.Drift, length=1),
])
line.particle_ref = xt.Particles(p0c=1.2e9, mass0=xt.PROTON_MASS_EV)

line_before_slicing = line.copy() # Keep for comparison

# Slice different elements with different strategies (in case multiple strategies
# apply to the same element, the last one takes precedence)
line.slice_thick_elements(
    slicing_strategies=[
        # Slicing with thin elements
        xt.Strategy(slicing=xt.Teapot(1)), # (1) Default applied to all elements
        xt.Strategy(slicing=xt.Uniform(2), element_type=xt.Bend), # (2) Selection by element type
        xt.Strategy(slicing=xt.Teapot(3), element_type=xt.Quadrupole),  # (4) Selection by element type
        xt.Strategy(slicing=xt.Teapot(4), name='mb1.*'), # (5) Selection by name pattern
        # Slicing with thick elements
        xt.Strategy(slicing=xt.Uniform(2, mode='thick'), name='mqf.*'), # (6) Selection by name pattern
        # Do not slice (leave untouched)
        xt.Strategy(slicing=None, name='mqd.1') # (7) Selection by name
    ])
line.build_tracker()
line_before_slicing.build_tracker()

# Ispect the result:

ltable = line.get_table(attr=True).cols['s', 'isthick', 'element_type',
                                        'parent_name', 'k0l', 'k1l', 'k2l']

# The sextupole msf.1 has one thin slice, as default strategy (1) is applied.
ltable.rows['msf.1_entry':'msf.1_exit']
# returns:
#
# Table: 5 rows, 8 cols
# name                s isthick element_type         parent_name k0l k1l   k2l
# msf.1_entry       0.3   False Marker                      None   0   0     0
# drift_msf.1..0    0.3    True DriftSliceSextupole        msf.1   0   0     0
# msf.1..0         0.35   False ThinSliceSextupole         msf.1   0   0 0.002
# drift_msf.1..1   0.35    True DriftSliceSextupole        msf.1   0   0     0
# msf.1_exit        0.4   False Marker                      None   0   0     0

# The bend mb2.1 has three thin slices, as strategy (2) is applied.
ltable.rows['mb2.1_entry':'mb2.1_exit']
# returns:
#
# Table: 9 rows, 8 cols
# name               s isthick element_type         parent_name      k0l k1l k2l
# mb2.1_entry      6.6   False Marker                      None        0   0   0
# mb2.1..entry_map 6.6   False ThinSliceBendEntry         mb2.1        0   0   0
# drift_mb2.1..0   6.6    True DriftSliceBend             mb2.1        0   0   0
# mb2.1..0         7.6   False ThinSliceBend              mb2.1 0.785398   0   0
# drift_mb2.1..1   7.6    True DriftSliceBend             mb2.1        0   0   0
# mb2.1..1         8.6   False ThinSliceBend              mb2.1 0.785398   0   0
# drift_mb2.1..2   8.6    True DriftSliceBend             mb2.1        0   0   0
# mb2.1..exit_map  9.6   False ThinSliceBendExit          mb2.1        0   0   0
# mb2.1_exit       9.6   False Marker                      None        0   0   0

# The quadrupole mqd.2 has four thin slices, as strategy (3) is applied.
ltable.rows['mqd.2_entry':'mqd.2_exit']
# returns:
#
# Table: 9 rows, 8 cols
# name                   s isthick element_type         parent_name k0l   k1l ...
# mqd.2_entry         15.9   False Marker                      None   0     0
# drift_mqd.2..0      15.9    True DriftSliceQuadrupole       mqd.2   0     0
# mqd.2..0         15.9375   False ThinSliceQuadrupole        mqd.2   0 -0.07
# drift_mqd.2..1   15.9375    True DriftSliceQuadrupole       mqd.2   0     0
# mqd.2..1           16.05   False ThinSliceQuadrupole        mqd.2   0 -0.07
# drift_mqd.2..2     16.05    True DriftSliceQuadrupole       mqd.2   0     0
# mqd.2..2         16.1625   False ThinSliceQuadrupole        mqd.2   0 -0.07
# drift_mqd.2..3   16.1625    True DriftSliceQuadrupole       mqd.2   0     0
# mqd.2_exit          16.2   False Marker                      None   0     0

# The quadrupole mqf.1 has two thick slices, as strategy (6) is applied.
ltable.rows['mqf.1_entry':'mqf.1_exit']
# returns:
#
# Table: 4 rows, 8 cols
# name                s isthick element_type         parent_name k0l   k1l k2l
# mqf.1_entry         0   False Marker                      None   0     0   0
# mqf.1..0            0    True ThickSliceQuadrupole       mqf.1   0 0.015   0
# mqf.1..1         0.15    True ThickSliceQuadrupole       mqf.1   0 0.015   0
# mqf.1_exit        0.3   False Marker                      None   0     0   0

# The quadrupole mqd.1 is left untouched, as strategy (7) is applied.
ltable.rows['mqd.1']
# returns:
#
# Table: 1 row, 8 cols
# name               s isthick element_type         parent_name k0l   k1l k2l
# mqd.1            5.3    True Quadrupole                  None   0 -0.21   0


########################################
# Change properties of sliced elements #
########################################

# Sliced elements are updated whenever their parent is changed. For example:

# Inspect a quadrupole:
ltable.rows['mqf.1.*']
# returns:
#
# Table: 4 rows, 8 cols
# name                s isthick element_type         parent_name k0l   k1l k2l
# mqf.1_entry         0   False Marker                      None   0     0   0
# mqf.1..0            0    True ThickSliceQuadrupole       mqf.1   0 0.015   0
# mqf.1..1         0.15    True ThickSliceQuadrupole       mqf.1   0 0.015   0
# mqf.1_exit        0.3   False Marker                      None   0     0   0

# Change the the strength of the parent
line['mqf.1'].k1 = 0.2

# Inspect
ltable = line.get_table(attr=True).cols['s', 'isthick', 'element_type',
                                        'parent_name', 'k0l', 'k1l', 'k2l']
ltable.rows['mqf.1.*']
# returns (the strength of the slices has changed):
#
# Table: 4 rows, 8 cols
# name                s isthick element_type         parent_name k0l  k1l k2l
# mqf.1_entry         0   False Marker                      None   0    0   0
# mqf.1..0            0    True ThickSliceQuadrupole       mqf.1   0 0.03   0
# mqf.1..1         0.15    True ThickSliceQuadrupole       mqf.1   0 0.03   0
# mqf.1_exit        0.3   False Marker                      None   0    0   0



import xtrack as xt
from cpymad.madx import Madx

# Load a very simple sequence from MAD-X
mad = Madx()
mad.input("""
    seq: sequence, l=4;
    b1: sbend, at=0.5, angle=0.2, l=1;
    q1: quadrupole, at=2.5, k1=0.1, l=1;
    endsequence;

    beam;
    use,sequence=seq;
""")

line = xt.Line.from_madx_sequence(mad.sequence.seq)
line.build_tracker()

print('The line as imported from MAD-X:')
print(line.get_table())

# Add multipolar components to elements
line['b1'].knl[2] = 0.001 # Normal sxtupole component
line['q1'].ksl[3] = 0.002 # Skew octupole component

tt = line.get_table(attr=True)
tt.cols['s', 'element_type', 'isthick', 'k2l', 'k3sl'].show()
# prints:
#
# name       s element_type isthick   k2l  k3sl
# seq$start  0 Marker         False     0     0
# b1         0 Bend            True 0.001     0
# drift_0    1 Drift           True     0     0
# q1         2 Quadrupole      True     0 0.002
# drift_1    3 Drift           True     0     0
# seq$end    4 Marker         False     0     0
# _end_point 4                False     0     0

# Slice the line
slicing_strategies = [
    xt.Strategy(slicing=None),  # Default catch-all
    xt.Strategy(slicing=xt.Teapot(2), element_type=xt.Bend),
    xt.Strategy(slicing=xt.Teapot(2), element_type=xt.Quadrupole),
]
line.slice_thick_elements(slicing_strategies)
line.build_tracker()

# Inspect
tt = line.get_table(attr=True)
tt.cols['s', 'element_type', 'isthick', 'parent_name', 'k2l', 'k3sl']

# returns:
#
# Table: 21 rows, 7 cols
# name                 s element_type         isthick parent_name    k2l  k3sl
# seq$start            0 Marker                 False        None      0     0
# b1_entry             0 Marker                 False        None      0     0
# b1..entry_map        0 ThinSliceBendEntry     False          b1      0     0
# drift_b1..0          0 DriftSliceBend          True          b1      0     0
# b1..0         0.166667 ThinSliceBend          False          b1 0.0005     0
# drift_b1..1   0.166667 DriftSliceBend          True          b1      0     0
# b1..1         0.833333 ThinSliceBend          False          b1 0.0005     0
# drift_b1..2   0.833333 DriftSliceBend          True          b1      0     0
# b1..exit_map         1 ThinSliceBendExit      False          b1      0     0
# b1_exit              1 Marker                 False        None      0     0
# drift_0              1 Drift                   True        None      0     0
# q1_entry             2 Marker                 False        None      0     0
# drift_q1..0          2 DriftSliceQuadrupole    True          q1      0     0
# q1..0          2.16667 ThinSliceQuadrupole    False          q1      0 0.001
# drift_q1..1    2.16667 DriftSliceQuadrupole    True          q1      0     0
# q1..1          2.83333 ThinSliceQuadrupole    False          q1      0 0.001
# drift_q1..2    2.83333 DriftSliceQuadrupole    True          q1      0     0
# q1_exit              3 Marker                 False        None      0     0
# drift_1              3 Drift                   True        None      0     0
# seq$end              4 Marker                 False        None      0     0

# Update misalignment for one element. We act on the parent and the effect is
# propagated to the slices.
line['q1'].knl[2] = -0.003
line['q1'].ksl[3] = -0.004

# Inspect
tt = line.get_table(attr=True)
tt.cols['s', 'element_type', 'isthick', 'parent_name', 'k2l', 'k3sl']
# returns:
#
# Table: 21 rows, 7 cols
# name                 s element_type         isthick parent_name     k2l   k3sl
# seq$start            0 Marker                 False        None       0      0
# b1_entry             0 Marker                 False        None       0      0
# b1..entry_map        0 ThinSliceBendEntry     False          b1       0      0
# drift_b1..0          0 DriftSliceBend          True          b1       0      0
# b1..0         0.166667 ThinSliceBend          False          b1  0.0005      0
# drift_b1..1   0.166667 DriftSliceBend          True          b1       0      0
# b1..1         0.833333 ThinSliceBend          False          b1  0.0005      0
# drift_b1..2   0.833333 DriftSliceBend          True          b1       0      0
# b1..exit_map         1 ThinSliceBendExit      False          b1       0      0
# b1_exit              1 Marker                 False        None       0      0
# drift_0              1 Drift                   True        None       0      0
# q1_entry             2 Marker                 False        None       0      0
# drift_q1..0          2 DriftSliceQuadrupole    True          q1       0      0
# q1..0          2.16667 ThinSliceQuadrupole    False          q1 -0.0015 -0.002
# drift_q1..1    2.16667 DriftSliceQuadrupole    True          q1       0      0
# q1..1          2.83333 ThinSliceQuadrupole    False          q1 -0.0015 -0.002
# drift_q1..2    2.83333 DriftSliceQuadrupole    True          q1       0      0
# q1_exit              3 Marker                 False        None       0      0
# drift_1              3 Drift                   True        None       0      0
# seq$end              4 Marker                 False        None       0      0
# _end_point           4                        False        None       0      0


import numpy as np

import numpy as np
import pandas as pd

import xtrack as xt
import xdeps as xd

import matplotlib.pyplot as plt

line = xt.Line.from_json('psb_01_with_chicane.json')
line.build_tracker()
line.twiss_default['method'] = '4d'

df = pd.read_csv('../../test_data/psb_chicane/chicane_collapse.csv',
                 delimiter=',', skipinitialspace=True)

line.functions['fun_bsw_k0l'] = xd.FunctionPieceWiseLinear(
    x=df['time'].values, y=df['bsw_k0l'].values)
line.functions['fun_bsw_k2l'] = xd.FunctionPieceWiseLinear(
    x=df['time'].values, y=df['bsw_k2l'].values)

# Control knob with function
line.vars['on_chicane_k0'] = 1
line.vars['on_chicane_k2'] = 1
line.vars['bsw_k0l'] = (line.functions.fun_bsw_k0l(line.vars['t_turn_s'])
                        * line.vars['on_chicane_k0'])
line.vars['bsw_k2l'] = (line.functions.fun_bsw_k2l(line.vars['t_turn_s'])
                        * line.vars['on_chicane_k2'])

line.to_json('psb_02_with_chicane_time_functions.json')

t_test = np.linspace(0, 6e-3, 100)

k0_bsw1 = []
k2l_bsw1 = []
k0_bsw2 = []
k2l_bsw2 = []
qx = []
qy = []
bety_at_qde3 = []
for ii, tt in enumerate(t_test):
    print(f'Twiss at t = {tt*1e3:.2f} ms   ', end='\r', flush=True)
    line.vars['t_turn_s'] = tt

    tw = line.twiss()

    qx.append(tw.qx)
    qy.append(tw.qy)
    bety_at_qde3.append(tw['bety', 'br.qde3'])
    k0_bsw1.append(line['bi1.bsw1l1.1'].k0)
    k2l_bsw1.append(line['bi1.bsw1l1.1'].knl[2])
    k0_bsw2.append(line['bi1.bsw1l1.2'].k0)
    k2l_bsw2.append(line['bi1.bsw1l1.2'].knl[2])

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
sp1 = plt.subplot(2,1,1)
plt.plot(t_test*1e3, k0_bsw1, label='k0 bsw1')
plt.plot(t_test*1e3, k0_bsw2, label='k0 bsw2')
plt.legend()
plt.subplot(2,1,2, sharex=sp1)
plt.plot(t_test*1e3, k2l_bsw1, label='k2l bsw1')
plt.plot(t_test*1e3, k2l_bsw2, label='k2l bsw2')
plt.legend()
plt.xlabel('time [ms]')

plt.figure(2)
sp1 = plt.subplot(2,1,1, sharex=sp1)
plt.plot(t_test*1e3, qy, label='qy')
plt.legend()
sp2 = plt.subplot(2,1,2, sharex=sp1)
plt.plot(t_test*1e3, bety_at_qde3, label='bety at qde3')
plt.legend()
plt.xlabel('time [ms]')

plt.show()


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json

import xtrack as xt
import xpart as xp
import xobjects as xo

context = xo.ContextCpu()

with open('../../test_data/hllhc15_noerrors_nobb/line_and_particle.json') as f:
    dct = json.load(f)
line = xt.Line.from_dict(dct['line'])
line.particle_ref = xt.Particles.from_dict(dct['particle'])

line.build_tracker()

num_particles = 50
particles = xp.generate_matched_gaussian_bunch(line=line,
                                               num_particles=num_particles,
                                               nemitt_x=2.5e-6,
                                               nemitt_y=2.5e-6,
                                               sigma_z=9e-2)

num_turns = 30
line.track(particles, num_turns=num_turns,
              turn_by_turn_monitor=True # <--
             )
# line.record_last_track contains the measured data. For example,
# line.record_last_track.x contains the x coordinate for all particles
# and all turns, e.g. line.record_last_track.x[3, 5] for the particle
# having particle_id = 3 and for the turn number 5.

# Monitor objects can be dumped to a dictionary and loaded back
mon = line.record_last_track
dct = mon.to_dict()
mon2 = xt.ParticlesMonitor.from_dict(dct)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

from . import elements as xstelems
from .base_classes import Element
from .be_beamfields import BeamBeam4D, BeamBeam6D, SCQGaussProfile

class_dict = {}
for nn in dir(xstelems):
    oo = getattr(xstelems, nn)
    if isinstance(oo, type):
        cc = oo
        if issubclass(cc, Element):
            class_dict[cc.__name__] = cc
class_dict['BeamBeam4D'] = BeamBeam4D
class_dict['BeamBeamBiGaussian2D'] = BeamBeam4D
class_dict['BeamBeam3D'] = BeamBeam6D
class_dict['BeamBeamBiGaussian3D'] = BeamBeam6D
class_dict['SpaceChargeBiGaussian'] = SCQGaussProfile



class TestLine:
    @classmethod
    def from_dict(cls, dct):

        self = cls(elements=[], element_names=[])

        if isinstance(dct['elements'], dict):
            el_dct_list = list(dct["elements"][nn] for nn in dct["element_names"])
        else:
            el_dct_list = dct["elements"]

        for el in el_dct_list:
            eltype = class_dict[el["__class__"]]
            eldct=el.copy()
            del eldct['__class__']
            newel = eltype.from_dict(eldct)
            self.elements.append(newel)
        self.element_names = dct["element_names"]
        return self

    def __init__(self, elements, element_names):
        self.elements = elements
        self.element_names = element_names

    def track(self, particles):
        for ee in self.elements:
            ee.track(particles)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import xtrack as xt

### Behavior:

# For num_turns = 1 (default):
#  - line.track(particles, ele_start=3, ele_stop=5) tracks once from element 3 to
#    element 5 e. particles.at_turn is not incremented as the particles never pass through
#    the end of the beam line.
#  - line.track(particles, ele_start=7, ele_stop=3) tracks once from element 5 to
#    the end of the beamline and then from element 0 to element 3 excluded.
#    particles.at_turn is incremented to one as the particles pass once
#    the end of the beam line.

# When indicating num_turns = N with N > 1, additional (N-1) full turns are added to logic
# above. Therefore:
# - if ele_stop < ele_start: stops at element ele_stop when particles.at_turn = N - 1
# - if ele_stop >= ele_start: stops at element ele_stop when particles.at_turn = N

elements = [xt.Drift(length=1.) for _ in range(10)]
line=xt.Line(elements=elements)
line.build_tracker()
line.reset_s_at_end_turn = False

# Standard mode
p = xt.Particles(x=[1e-3, 2e-3, 3e-3], p0c=7e12)
line.track(p, num_turns=4)
p.at_turn # is [4, 4, 4]

# ele_start > 0
p = xt.Particles(x=[1e-3, 2e-3, 3e-3], p0c=7e12, s=4., at_element=2)
line.track(p, num_turns=3, ele_start=2)
p.at_turn # is [3, 3, 3]

# 0 <= ele_start < ele_stop
p = xt.Particles(x=[1e-3, 2e-3, 3e-3], p0c=7e12, s=5 * 2., at_element=5)
line.track(p, num_turns=4, ele_start=5, ele_stop=8)
p.at_turn # is [3, 3, 3]

# 0 <= ele_start < ele_stop -- num_turns = 1
p = xt.Particles(x=[1e-3, 2e-3, 3e-3], p0c=7e12, s=5 * 2., at_element=5)
line.track(p, num_turns=1, ele_start=5, ele_stop=2)
p.at_turn # is [1, 1, 1]

# 0 <= ele_start < ele_stop -- num_turns > 1
p = xt.Particles(x=[1e-3, 2e-3, 3e-3], p0c=7e12, s=5 * 2., at_element=5)
line.track(p, num_turns=4, ele_start=5, ele_stop=2)
p.at_turn # is [4, 4, 4]

# Use ele_start and num_turns
p = xt.Particles(x=[1e-3, 2e-3, 3e-3], p0c=7e12, s=5 * 2., at_element=5)
line.track(p, ele_start=5, num_elements=7)
p.at_turn # is [1, 1, 1]





import numpy as np
import xtrack as xt
import xobjects as xo

num_particles_test = 300

fname = 'fccee_z'; gemitt_y_target = 1.4e-12; n_turns_track_test = 3000
# fname = 'fccee_w'; gemitt_y_target = 2.2e-12; n_turns_track_test = 2000
# fname = 'fccee_h'; gemitt_y_target = 1.4e-12; n_turns_track_test = 400
fname = 'fccee_t'; gemitt_y_target = 2e-12; n_turns_track_test = 600

line = xt.Line.from_json(fname + '_thin.json')

# Add monitor in a dispersion-free place out of crab waist
monitor_at = 'qrdr2.3_entry@start'
monitor = xt.ParticlesMonitor(num_particles=num_particles_test,
                              start_at_turn=0, stop_at_turn=n_turns_track_test)
line.insert('monitor', monitor, at='qrdr2.3_entry@start')

line.build_tracker()

tw_no_rad = line.twiss(method='4d')

line.configure_radiation(model='mean')
line.compensate_radiation_energy_loss()

tw_rad_wig_off = line.twiss(eneloss_and_damping=True)

line.vars['on_wiggler_v'] = 0.05
line.compensate_radiation_energy_loss()
opt = line.match(
    solve=False,
    eneloss_and_damping=True,
    compensate_radiation_energy_loss=True,
    targets=[
        xt.Target(eq_gemitt_y=gemitt_y_target, tol=1e-15, optimize_log=True)],
    vary=xt.Vary('on_wiggler_v', step=0.01, limits=(0.02, 2))
)

opt.solve()

tw_rad = line.twiss(eneloss_and_damping=True)
tw_rad2 = line.twiss(eneloss_and_damping=True, method='6d',
                     radiation_method='full',
                     compute_lattice_functions=False,
                     compute_chromatic_properties=False)


ex = tw_rad.eq_gemitt_x
ey = tw_rad.eq_gemitt_y
ez = tw_rad.eq_gemitt_zeta

line.configure_radiation(model='quantum')
p = line.build_particles(num_particles=num_particles_test)

# Switch to multithreaded
line.discard_tracker()
line.build_tracker(_context=xo.ContextCpu(omp_num_threads='auto'), use_prebuilt_kernels=False)

line.track(p, num_turns=n_turns_track_test, turn_by_turn_monitor=True, time=True,
           with_progress=10)
mon_at_start = line.record_last_track
print(f'Tracking time: {line.time_last_track}')

line.configure_radiation(model='mean')

import matplotlib.pyplot as plt
plt.close('all')
for ii, (mon, element_mon, label) in enumerate(
                            [(mon_at_start, 0, 'inside crab waist'),
                             (monitor, 'monitor', 'outside crab waist')
                             ]):

    betx = tw_rad['betx', element_mon]
    bety = tw_rad['bety', element_mon]
    betx2 = tw_rad['betx2', element_mon]
    bety1 = tw_rad['bety1', element_mon]
    dx = tw_rad['dx', element_mon]
    dy = tw_rad['dy', element_mon]

    fig = plt.figure(ii + 1, figsize=(6.4, 4.8*1.3))
    spx = fig. add_subplot(3, 1, 1)
    spx.plot(np.std(mon.x, axis=0), label='track')
    spx.axhline(
        tw_rad2.eq_beam_covariance_matrix['sigma_x', element_mon],
        color='green', label='twiss')
    spx.axhline(
        np.sqrt(ex * betx + ey * betx2 + (np.std(p.delta) * dx)**2),
        color='red', label='twiss')
    spx.legend(loc='lower right')
    spx.set_ylabel(r'$\sigma_{x}$ [m]')
    spx.set_ylim(bottom=0)

    spy = fig. add_subplot(3, 1, 2, sharex=spx)
    spy.plot(np.std(mon.y, axis=0), label='track')
    spy.axhline(
        tw_rad2.eq_beam_covariance_matrix['sigma_y', element_mon],
        color='green', label='twiss')
    spy.axhline(
        np.sqrt(ex * bety1 + ey * bety + (np.std(p.delta) * dy)**2),
        color='red', label='twiss')
    spy.set_ylabel(r'$\sigma_{y}$ [m]')
    spy.set_ylim(bottom=0)

    spz = fig. add_subplot(3, 1, 3, sharex=spx)
    spz.plot(np.std(mon.zeta, axis=0))
    spz.axhline(
        tw_rad2.eq_beam_covariance_matrix['sigma_zeta', element_mon],
        color='green')
    spz.axhline(np.sqrt(ez * tw_rad.bets0), color='red')
    spz.set_ylabel(r'$\sigma_{z}$ [m]')
    spz.set_ylim(bottom=0)

    plt.suptitle(f'{fname} - ' r'$\varepsilon_y$ = ' f'{ey*1e12:.6f} pm - {label}')

plt.show()


import json
import numpy as np
from cpymad.madx import Madx

import xtrack as xt
import xdeps as xd
import xobjects as xo

import matplotlib.pyplot as plt

bumper_names = ['bi1.bsw1l1.1', 'bi1.bsw1l1.2', 'bi1.bsw1l1.3', 'bi1.bsw1l1.4']
thick_bumpers = {
'bi1.bsw1l1.1' : {'k0_name': 'k0BI1BSW1L11'},
'bi1.bsw1l1.2' : {'k0_name': 'k0BI1BSW1L12'},
'bi1.bsw1l1.3' : {'k0_name': 'k0BI1BSW1L13'},
'bi1.bsw1l1.4' : {'k0_name': 'k0BI1BSW1L14'},
}

mad = Madx()

# Load model
mad.input('''
call, file = '../../test_data/psb_chicane/psb.seq';
call, file = '../../test_data/psb_chicane/psb_fb_lhc.str';
beam, particle=PROTON, pc=0.5708301551893517;
use, sequence=psb1;
twiss;
''')

# Add shifts
mad.input('''
SELECT,FLAG=ERROR,CLEAR;
SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.1*;
EALIGN, DX=-0.0057;

SELECT,FLAG=ERROR,CLEAR;
SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.2*;
SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.3*;
SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.4*;
EALIGN, DX=-0.0442;
''')

# Match tunes
mad.input(f'''
QH = 4.4;
QV = 4.45;

MATCH, Sequence=psb1;
    VARY, NAME = kbrqf, STEP = 1e-3;
    VARY, NAME = kbrqd, STEP = 1e-3;
    GLOBAL, Q1 = QH;
    GLOBAL, Q2 = QV;
    JACOBIAN,CALLS=1000,TOLERANCE=1.0E-18,STRATEGY=3;
ENDMATCH;
''')

# # Handle bumpers (angle = 0)
# seq = mad.sequence.psb1
# seq.expanded_elements['bi1.bsw1l1.1'].angle = 1e-20
# seq.expanded_elements['bi1.bsw1l1.2'].angle = 1e-20
# seq.expanded_elements['bi1.bsw1l1.3'].angle = 1e-20
# seq.expanded_elements['bi1.bsw1l1.4'].angle = 1e-20

# Store bumpers length
for nn in bumper_names:
    thick_bumpers[nn]['length'] = mad.sequence.psb1.expanded_elements[nn].l

# Set K0 and K2 for thick bumpers
mad.input(f'''

l_bsw1l1.1 = {thick_bumpers['bi1.bsw1l1.1']['length']};
l_bsw1l1.2 = {thick_bumpers['bi1.bsw1l1.2']['length']};
l_bsw1l1.3 = {thick_bumpers['bi1.bsw1l1.3']['length']};
l_bsw1l1.4 = {thick_bumpers['bi1.bsw1l1.4']['length']};

BSW_K0L = 6.6E-2;
BSW_K2L = -9.7429e-02;
'''
)

mad.input('''
SELECT,FLAG=ERROR,CLEAR;
SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.1;
SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.4;
EFCOMP, DKN:={+BSW_K0L, 0, +BSW_K2L};
!EFCOMP, DKN:={0, 0, +BSW_K2L};

SELECT,FLAG=ERROR,CLEAR;
SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.2;
SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.3;
EFCOMP, DKN:={-BSW_K0L, 0, -BSW_K2L};

''')

# Twiss and ptc twiss with errors
twmad = mad.twiss()

beta0 = mad.sequence.psb1.beam.beta

twmad = twmad.dframe()

mad.input('''
ptc_create_universe;
ptc_create_layout, model=3, method=6, nst=5, exact=true;
ptc_setswitch, debuglevel=0, nocavity=false, fringe=true,
            exact_mis=true, time=true, totalpath=true;
select_ptc_normal, q1=0, q2=0;
select_ptc_normal, dq1=1, dq2=1;
select_ptc_normal, dq1=2, dq2=2;
PTC_ALIGN;
ptc_normal, closed_orbit, normal, icase=5, no=3;
ptc_twiss, closed_orbit, table=ptc_twiss, icase=4, no=3,
            summary_table=ptc_twiss_summary;
ptc_end;
''')
twptc = mad.table.ptc_twiss.dframe()
qx_ptc = mad.table.ptc_twiss.mu1[-1]
qy_ptc = mad.table.ptc_twiss.mu2[-1]
dqx_ptc = mad.table.normal_results.value[2] * beta0
dqy_ptc = mad.table.normal_results.value[3] * beta0
t_ptc = xd.Table(mad.table.ptc_twiss)

line_thick = xt.Line.from_json('psb_03_with_chicane_corrected.json')
line_thick.build_tracker()
line_thick.configure_bend_model(core='full', edge='full')
line_thick.vars['on_chicane_beta_corr'] = 0
line_thick.vars['on_chicane_tune_corr'] = 0

line_thin = xt.Line.from_json('psb_04_with_chicane_corrected_thin.json')
line_thin.build_tracker()
line_thin.vars['on_chicane_beta_corr'] = 0
line_thin.vars['on_chicane_tune_corr'] = 0

t_test = np.linspace(0, 6e-3, 100)

qx_thick = []
qy_thick = []
dqx_thick = []
dqy_thick = []
bety_at_scraper_thick = []
qx_thin = []
qy_thin = []
dqx_thin = []
dqy_thin = []
bety_at_scraper_thin = []

qx_ptc = []
qy_ptc = []
dqx_ptc = []
dqy_ptc = []
bety_at_scraper_ptc = []
for ii, tt in enumerate(t_test):
    print(f'Twiss at t = {tt*1e3:.2f} ms   ', end='\r', flush=True)
    line_thick.vars['t_turn_s'] = tt
    line_thin.vars['t_turn_s'] = tt

    tw_thick = line_thick.twiss()
    bety_at_scraper_thick.append(tw_thick['bety', 'br.stscrap22'])
    qx_thick.append(tw_thick.qx)
    qy_thick.append(tw_thick.qy)
    dqx_thick.append(tw_thick.dqx)
    dqy_thick.append(tw_thick.dqy)

    tw_thin = line_thin.twiss()
    bety_at_scraper_thin.append(tw_thin['bety', 'br.stscrap22'])
    qx_thin.append(tw_thin.qx)
    qy_thin.append(tw_thin.qy)
    dqx_thin.append(tw_thin.dqx)
    dqy_thin.append(tw_thin.dqy)

    mad.globals.bsw_k0l = line_thick.vars['bsw_k0l']._value
    mad.globals.bsw_k2l = line_thick.vars['bsw_k2l']._value

    mad.input('''
    SELECT,FLAG=ERROR,CLEAR;
    SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.1;
    SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.4;
    EFCOMP, DKN:={+BSW_K0L, 0, +BSW_K2L};
    !EFCOMP, DKN:={0, 0, +BSW_K2L};

    SELECT,FLAG=ERROR,CLEAR;
    SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.2;
    SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.3;
    EFCOMP, DKN:={-BSW_K0L, 0, -BSW_K2L};

    ''')

    mad.input('''
        ptc_create_universe;
        ptc_create_layout, model=3, method=6, nst=5, exact=true;
        ptc_setswitch, debuglevel=0, nocavity=false, fringe=true,
                    exact_mis=true, time=true, totalpath=true;
        select_ptc_normal, q1=0, q2=0;
        select_ptc_normal, dq1=1, dq2=1;
        select_ptc_normal, dq1=2, dq2=2;
        PTC_ALIGN;
        ptc_normal, closed_orbit, normal, icase=5, no=3;
        ptc_twiss, closed_orbit, table=ptc_twiss, icase=4, no=3,
                    summary_table=ptc_twiss_summary;
        ptc_end;
        ''')

    qx_ptc.append(mad.table.ptc_twiss.mu1[-1])
    qy_ptc.append(mad.table.ptc_twiss.mu2[-1])
    dqx_ptc.append(mad.table.normal_results.value[2] * beta0)
    dqy_ptc.append(mad.table.normal_results.value[3] * beta0)

    tw_ptc = xd.Table(mad.table.ptc_twiss)

    bety_at_scraper_ptc.append(tw_ptc['bety', 'br.stscrap22:1'])

qx_thick = np.array(qx_thick)
qy_thick = np.array(qy_thick)
dqx_thick = np.array(dqx_thick)
dqy_thick = np.array(dqy_thick)
bety_at_scraper_thick = np.array(bety_at_scraper_thick)
qx_thin = np.array(qx_thin)
qy_thin = np.array(qy_thin)
dqx_thin = np.array(dqx_thin)
dqy_thin = np.array(dqy_thin)
bety_at_scraper_thin = np.array(bety_at_scraper_thin)
qx_ptc = np.array(qx_ptc)
qy_ptc = np.array(qy_ptc)
dqx_ptc = np.array(dqx_ptc)
dqy_ptc = np.array(dqy_ptc)
bety_at_scraper_ptc = np.array(bety_at_scraper_ptc)

assert np.allclose(qx_thick, qx_ptc, atol=2e-4, rtol=0)
assert np.allclose(qy_thick, qy_ptc, atol=2e-4, rtol=0)
assert np.allclose(qx_thin, qx_ptc, atol=1e-3, rtol=0)
assert np.allclose(qy_thin, qy_ptc, atol=1e-3, rtol=0)
assert np.allclose(dqx_thick, dqx_ptc, atol=0.5, rtol=0)
assert np.allclose(dqy_thick, dqy_ptc, atol=0.5, rtol=0)
assert np.allclose(dqx_thin, dqx_ptc, atol=0.5, rtol=0)
assert np.allclose(dqy_thin, dqy_ptc, atol=0.5, rtol=0)
assert np.allclose(bety_at_scraper_thick, bety_at_scraper_ptc, atol=0, rtol=1e-2)
assert np.allclose(bety_at_scraper_thin, bety_at_scraper_ptc, atol=0, rtol=2e-2)

with open('ptc_ref.json', 'w') as fid:
    json.dump({
        't_test': t_test,
        'qx_ptc': qx_ptc,
        'qy_ptc': qy_ptc,
        'dqx_ptc': dqx_ptc,
        'dqy_ptc': dqy_ptc,
        'bety_at_scraper_ptc': bety_at_scraper_ptc,
    }, fid, cls=xo.JEncoder)




import matplotlib.pyplot as plt
plt.close('all')

fig1 = plt.figure(1)
sp1 = plt.subplot(2,1,1)

plt.plot(t_test*1e3, qy_thick, '-', color='C0', label='xsuite thick')
plt.plot(t_test*1e3, qy_thin, '-.', color='C1', label='xsuite thin')
plt.plot(t_test*1e3, qy_ptc, '--', color='C2', label='ptc')
plt.ylabel('tune')


plt.legend()

sp2 = plt.subplot(2,1,2, sharex=sp1)
plt.plot(t_test*1e3, bety_at_scraper_thick, '-', color='C0', label='xsuite thick')
plt.plot(t_test*1e3, bety_at_scraper_thin, '-.', label='xsuite thin')
plt.plot(t_test*1e3, bety_at_scraper_ptc, '--',label='ptc')
plt.ylim(bottom=0)

plt.legend()

plt.xlabel('time [ms]')
plt.ylabel(r'$\beta_y$ at scraper [m]')

plt.figure(2)
sp1 = plt.subplot(2,1,1, sharex=sp1)
plt.plot(t_test*1e3, dqx_thick, '-', color='C0', label='xsuite thick')
plt.plot(t_test*1e3, dqx_thin, '-.', color='C1', label='xsuite thin')
plt.plot(t_test*1e3, dqx_ptc, '--', color='C2', label='ptc')
plt.ylim(bottom=-9, top=0)
plt.ylabel(r"$Q'_x$")
plt.legend()

sp2 = plt.subplot(2,1,2, sharex=sp1)
plt.plot(t_test*1e3, dqy_thick, '-', color='C0', label='xsuite thick')
plt.plot(t_test*1e3, dqy_thin, '-.', color='C1', label='xsuite thin')
plt.plot(t_test*1e3, dqy_ptc, '--', color='C2', label='ptc')
plt.ylabel(r"$Q'_y$")
plt.ylim(bottom=-9,top=0)
plt.legend()

plt.xlabel('time [ms]')

plt.show()




import xtrack as xt
import xobjects as xo
import numpy as np

assert_allclose = xo.assert_allclose

elements = {
    'd1': xt.Drift(length=1),
    'm1': xt.Marker(),
    'd2': xt.Drift(length=1),
}

line=xt.Line(elements=elements,
             element_names=list(elements.keys()))

# Note that the name is reused
line.insert_element(element=xt.Bend(length=1.), name='m1', at_s=0.5)

tt = line.get_table()

assert np.all(tt.name == ['d1..0', 'm1', 'd2..1', '_end_point'])
assert np.all(tt.parent_name == ['d1', None, 'd2', None])
assert_allclose(tt.s, [0. , 0.5, 1.5, 2. ], rtol=0, atol=1e-14)


import numpy as np
import xtrack as xt

pi = np.pi
lbend = 3

# Build a simple ring
env = xt.Environment()
line = env.new_line(components=[
    env.new('mqf.1', xt.Quadrupole, length=0.3, k1=0.1),
    env.new('d1.1',  xt.Drift, length=1),
    env.new('mb1.1', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d2.1',  xt.Drift, length=1),

    env.new('mqd.1', xt.Quadrupole, length=0.3, k1=-0.7),
    env.new('d3.1',  xt.Drift, length=1),
    env.new('mb2.1', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d4.1',  xt.Drift, length=1),

    env.new('mqf.2', xt.Quadrupole, length=0.3, k1=0.1),
    env.new('d1.2',  xt.Drift, length=1),
    env.new('mb1.2', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d2.2',  xt.Drift, length=1),

    env.new('mqd.2', xt.Quadrupole, length=0.3, k1=-0.7),
    env.new('d3.2',  xt.Drift, length=1),
    env.new('mb2.2', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d4.2',  xt.Drift, length=1),
])
line.particle_ref = xt.Particles(p0c=1.2e9, mass0=xt.PROTON_MASS_EV)

# Quick access to an element and its attributes (by name)
line['mqf.1'] # is Quadrupole(length=0.3, k1=0.1, ...)
line['mqf.1'].k1 # is 0.1
line['mqf.1'].length # is 0.3

# Quick access to an element and its attributes (by index)
line[0] # is Quadrupole(length=0.3, k1=0.1, ...)
line[0].k1 # is 0.1
line[0].length # is 0.3

# Tuple with all element names
line.element_names # is ('mqf.1', 'd1.1', 'mb1.1', 'd2.1', 'mqd.1', ...

# Tuple with all element objects
line.elements # is (Quadrupole(length=0.3, k1=0.1, ...), Drift(length=1), ...

# `line.attr[...]` can be used for efficient extraction of a given attribute for
# all elements. For example:
line.attr['length'] # is (0.3, 1, 3, 1, 0.3, 1, 3, 1, 0.3, 1, 3, 1, 0.3, 1, 3, 1)
line.attr['k1l'] # is ('0.03, 0.0, 0.0, 0.0, -0.21, 0.0, 0.0, 0.0, 0.03, ... )

# The list of all attributes can be found in
line.attr.keys() # is ('length', 'k1', 'k1l', 'k2', 'k2l', 'k3', 'k3l', 'k4', ...

# `line.get_table()`` can be used to get a table with information about the line
# elements. For example:
tab = line.get_table()

# The table can be printed
tab.show()
# prints:
#
# name          s element_type isthick isreplica parent_name iscollective
# mqf.1         0 Quadrupole      True     False        None        False
# d1.1        0.3 Drift           True     False        None        False
# mb1.1       1.3 Bend            True     False        None        False
# d2.1        4.3 Drift           True     False        None        False
# mqd.1       5.3 Quadrupole      True     False        None        False
# d3.1        5.6 Drift           True     False        None        False
# mb2.1       6.6 Bend            True     False        None        False
# d4.1        9.6 Drift           True     False        None        False
# mqf.2      10.6 Quadrupole      True     False        None        False
# d1.2       10.9 Drift           True     False        None        False
# mb1.2      11.9 Bend            True     False        None        False
# d2.2       14.9 Drift           True     False        None        False
# mqd.2      15.9 Quadrupole      True     False        None        False
# d3.2       16.2 Drift           True     False        None        False
# mb2.2      17.2 Bend            True     False        None        False
# d4.2       20.2 Drift           True     False        None        False
# _end_point 21.2                False     False        None        False

# Access to a single element of the table
tab['s', 'mb2.1'] # is 6.6

# Access to a single column of the table
tab['s'] # is [0.0, 0.3, 1.3, 4.3, 5.3, 5.6, 6.6, 9.6, 10.6, 10.9, 11.9, ...

# Regular expressions can be used to select elements by name
tab.rows['mb.*']
# returns:
#
# Table: 4 rows, 94 cols
# name          s element_type isthick isreplica parent_name iscollective
# mb1.1       1.3 Bend            True     False        None        False
# mb2.1       6.6 Bend            True     False        None        False
# mb1.2      11.9 Bend            True     False        None        False
# mb2.2      17.2 Bend            True     False        None        False


# Elements can be selected by type
tab.rows[tab.element_type == 'Quadrupole']
# returns:
#
# Table: 4 rows, 94 cols
# name          s element_type isthick isreplica parent_name iscollective
# mqf.1         0 Quadrupole      True     False        None        False
# mqd.1       5.3 Quadrupole      True     False        None        False
# mqf.2      10.6 Quadrupole      True     False        None        False
# mqd.2      15.9 Quadrupole      True     False        None        False

# A section of the ring can be selected using names
tab.rows['mqd.1':'mqd.2']
# returns:
#
# Table: 9 rows, 94 cols
# name          s element_type isthick isreplica parent_name iscollective
# mqd.1       5.3 Quadrupole      True     False        None        False
# d3.1        5.6 Drift           True     False        None        False
# mb2.1       6.6 Bend            True     False        None        False
# d4.1        9.6 Drift           True     False        None        False
# mqf.2      10.6 Quadrupole      True     False        None        False
# d1.2       10.9 Drift           True     False        None        False
# mb1.2      11.9 Bend            True     False        None        False
# d2.2       14.9 Drift           True     False        None        False
# mqd.2      15.9 Quadrupole      True     False        None        False

# A section of the ring can be selected using the s coordinate
tab.rows[3.0:7.0:'s']
# returns:
#
# Table: 4 rows, 94 cols
# name         s element_type isthick isreplica parent_name iscollective
# d2.1       4.3 Drift           True     False        None        False
# mqd.1      5.3 Quadrupole      True     False        None        False
# d3.1       5.6 Drift           True     False        None        False
# mb2.1      6.6 Bend            True     False        None        False

# A section of the ring can be selected using indexes relative one element
# (e.g. to get from three elements upstream of 'mqd.1' to two elements
# downstream of 'mb2.1')
tab.rows['mqd.1<<3':'mb2.1>>2']
# returns:
#
# Table: 8 rows, 94 cols
# name          s element_type isthick isreplica parent_name iscollective
# d1.1        0.3 Drift           True     False        None        False
# mb1.1       1.3 Bend            True     False        None        False
# d2.1        4.3 Drift           True     False        None        False
# mqd.1       5.3 Quadrupole      True     False        None        False
# d3.1        5.6 Drift           True     False        None        False
# mb2.1       6.6 Bend            True     False        None        False
# d4.1        9.6 Drift           True     False        None        False
# mqf.2      10.6 Quadrupole      True     False        None        False

# Each of the selection methods above returns a valid table, hence selections
# can be chained. For example:
tab.rows[0:10:'s'].rows['mb.*']
# returns:
#
# Table: 2 rows, 94 cols
# name         s element_type isthick isreplica parent_name iscollective
# mb1.1      1.3 Bend            True     False        None        False
# mb2.1      6.6 Bend            True     False        None        False

# or more efficiently
tab.rows[0:10:'s','mb.*']
# returns:
#
# Table: 2 rows, 94 cols
# name         s element_type isthick isreplica parent_name iscollective
# mb1.1      1.3 Bend            True     False        None        False
# mb2.1      6.6 Bend            True     False        None        False


# All attributes extracted by `line.attr[...]` can be included in the table
# using `attr=True`. For example, using `tab.cols[...]` to select columns, we
# we can get the focusing strength of all quadrupoles in the ring:
tab = line.get_table(attr=True)
tab.rows[tab.element_type=='Quadrupole'].cols['s length k1l']
# returns:
#
# Table: 4 rows, 4 cols
# name          s length   k1l
# mqf.1         0    0.3  0.03
# mqd.1       5.3    0.3 -0.21
# mqf.2      10.6    0.3  0.03
# mqd.2      15.9    0.3 -0.21



import json
import numpy as np
import xtrack as xt

#########################################
# Load line and twiss with no radiation #
#########################################

filename = '../../test_data/clic_dr/line_for_taper.json'

with open(filename, 'r') as f:
    line = xt.Line.from_dict(json.load(f))
line.build_tracker()

line.configure_radiation(model=None)
tw_no_rad = line.twiss(method='4d', freeze_longitudinal=True)

###############################################
# Enable radiation and compensate energy loss #
###############################################

line.configure_radiation(model='mean')

# - Set cavity lags to compensate energy loss
# - Taper magnet strengths to avoid optics and orbit distortions
line.compensate_radiation_energy_loss()

##############################
# Twiss to check the results #
##############################

tw = line.twiss(method='6d')

tw.delta # contains the momentum deviation along the ring

#!end-doc-part

import matplotlib.pyplot as plt
plt.close('all')

fig1 = plt.figure(1, figsize=(6.4, 4.8*1.5))
spbet = plt.subplot(3,1,1)
spco = plt.subplot(3,1,2, sharex=spbet)
spdisp = plt.subplot(3,1,3, sharex=spbet)

spbet.plot(tw['s'], tw['betx'])
spbet.plot(tw['s'], tw['bety'])

spco.plot(tw['s'], tw['x'])
spco.plot(tw['s'], tw['y'])

spdisp.plot(tw['s'], tw['dx'])
spdisp.plot(tw['s'], tw['dy'])

spbet.set_ylabel(r'$\beta_{x,y}$ [m]')
spco.set_ylabel(r'(Closed orbit)$_{x,y}$ [m]')
spdisp.set_ylabel(r'$D_{x,y}$ [m]')
spdisp.set_xlabel('s [m]')

fig1.suptitle(
    r'$q_x$ = ' f'{tw["qx"]:.5f}' r' $q_y$ = ' f'{tw["qy"]:.5f}' '\n'
    r"$Q'_x$ = " f'{tw["dqx"]:.2f}' r" $Q'_y$ = " f'{tw["dqy"]:.2f}'
    r' $\gamma_{tr}$ = '  f'{1/np.sqrt(tw["momentum_compaction_factor"]):.2f}'
)

fig1.subplots_adjust(left=.15, right=.92, hspace=.27)
plt.show()


import numpy as np
import xtrack as xt

d_zeta = 1e-3

collider = xt.Environment.from_json(
    '../../test_data/hllhc15_collider/collider_00_from_mad.json'
)
collider.build_trackers()

collider.vars['vrf400'] = 16
collider.vars['on_crab1'] = -190
collider.vars['on_crab5'] = -190

line = collider.lhcb1

tw6d_rf_on = line.twiss()
tw4d_rf_on = line.twiss(method='4d')

collider.vars['vrf400'] = 0
tw4d_rf_off = line.twiss(method='4d')

collider.vars['vrf400'] = 16
collider.vars['on_crab1'] = 0
collider.vars['on_crab5'] = 0

line = collider.lhcb1

tw6d_rf_on_crab_off = line.twiss()
tw4d_rf_on_crab_off = line.twiss(method='4d')

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
sp1 = plt.subplot(1,1,1)
plt.plot(tw6d_rf_on.s, tw6d_rf_on.dx_zeta, label='6D - RF on')
plt.plot(tw4d_rf_on.s, tw4d_rf_on.dx_zeta, '--', label='4D - RF on')
plt.plot(tw4d_rf_off.s, tw4d_rf_off.dx_zeta, '-.', label='4D - RF off')
plt.legend()
plt.xlabel('s [m]')
plt.ylabel('dx/dzeta')

plt.figure(2)
sp2 = plt.subplot(1,1,1, sharex=sp1, sharey=sp1)
plt.plot(tw6d_rf_on_crab_off.s, tw6d_rf_on_crab_off.dx_zeta, label='6D - RF on, crab off')
plt.plot(tw4d_rf_on_crab_off.s, tw4d_rf_on_crab_off.dx_zeta, '--', label='4D - RF on, crab off')
plt.legend()
plt.xlabel('s [m]')
plt.ylabel('dx/dzeta')
plt.show()






import xtrack as xt

collider = xt.Environment.from_json(
    '../../test_data/hllhc15_thick/hllhc15_collider_thick.json')
collider['lhcb1'].twiss_default['method'] = '4d'
collider['lhcb2'].twiss_default['method'] = '4d'

line = collider.lhcb1

tw = line.twiss()

init = xt.TwissInit(betx=0.15, bety=0.15,
                           ax_chrom=42.7928, bx_chrom=-18.4181,
                           ay_chrom=-18.0191, by_chrom= 11.54862)
tw_open = line.twiss(start='ip5', end='ip7', init=init,
                        compute_chromatic_properties=True)

import matplotlib.pyplot as plt
plt.close('all')

plt.figure(1, figsize=(6.4, 4.8*1.5))
ax1 = plt.subplot(4, 1, 1)
plt.plot(tw.s, tw.ax_chrom, label='closed')
plt.plot(tw_open.s, tw_open.ax_chrom, label='open')
plt.ylabel(r'$A_x$')
plt.legend()

plt.subplot(4, 1, 2, sharex=ax1)
plt.plot(tw.s, tw.bx_chrom, label='closed')
plt.plot(tw_open.s, tw_open.bx_chrom, label='open')
plt.ylabel(r'$B_x$')

plt.subplot(4, 1, 3, sharex=ax1)
plt.plot(tw.s, tw.ay_chrom, label='closed')
plt.plot(tw_open.s, tw_open.ay_chrom, label='open')
plt.ylabel(r'$A_y$')

plt.subplot(4, 1, 4, sharex=ax1)
plt.plot(tw.s, tw.by_chrom, label='closed')
plt.plot(tw_open.s, tw_open.by_chrom, label='open')
plt.ylabel(r'$B_y$')

plt.xlabel('s [m]')

plt.figure(2, figsize=(6.4, 4.8))
plt.subplot(2, 1, 1, sharex=ax1)
plt.plot(tw.s, tw.wx_chrom, label='closed')
plt.plot(tw_open.s, tw_open.wx_chrom, label='open')
plt.ylabel(r'$W_x$')

plt.subplot(2, 1, 2, sharex=ax1)
plt.plot(tw.s, tw.wy_chrom, label='closed')
plt.plot(tw_open.s, tw_open.wy_chrom, label='open')
plt.ylabel(r'$W_y$')

plt.legend


plt.show()



import xtrack as xt
import xobjects as xo

magnet = xt.Magnet(k0=0.002, h=0.002, k1=0.02, length=2)

m_exact = magnet.copy()
m_exact.model = 'bend-kick-bend'
m_exact.integrator='yoshida4'
m_exact.num_multipole_kicks = 1000

m_expanded = magnet.copy()
m_expanded.model = 'mat-kick-mat'
m_expanded.integrator='yoshida4'
m_expanded.num_multipole_kicks = 1000

p0 = xt.Particles(x=1e-3, y=2e-3, px=5e-6)

p_exact = p0.copy()
m_exact.track(p_exact)

p_expanded = p0.copy()
m_expanded.track(p_expanded)

xo.assert_allclose(p_exact.x, p_expanded.x, rtol=0, atol=1e-10)
xo.assert_allclose(p_exact.px, p_expanded.px, rtol=0, atol=1e-11)
xo.assert_allclose(p_exact.y, p_expanded.y, rtol=0, atol=2e-10)
xo.assert_allclose(p_exact.py, p_expanded.py, rtol=0, atol=1e-11)
xo.assert_allclose(p_exact.zeta, p_expanded.zeta, rtol=0, atol=1e-12)
xo.assert_allclose(p_exact.delta, p_expanded.delta, rtol=0, atol=1e-14)








# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json
import pathlib

import numpy as np
import pytest

import xfields as xf
import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts, fix_random_seed


@for_all_test_contexts
@pytest.mark.parametrize(
    'mode',
    ['frozen', 'quasi-frozen', 'pic', 'pic_average_transverse'],
)
@fix_random_seed(746483)
def test_ring_with_spacecharge(test_context, mode):

    test_data_folder = pathlib.Path(
            __file__).parent.joinpath('../test_data').absolute()
    fname_line = test_data_folder.joinpath('sps_w_spacecharge/'
                      'line_no_spacecharge_and_particle.json')

    # Test settings (fast but inaccurate)
    bunch_intensity = 1e11/3  # Need short bunch to avoid bucket non-linearity
    sigma_z = 22.5e-2/3
    nemitt_x = 2.5e-6
    nemitt_y = 2.5e-6
    n_part = int(1e6/10)*10
    nz_grid = 100//20
    z_range = (-3*sigma_z/40, 3*sigma_z/40)

    num_spacecharge_interactions = 540

    ##############
    # Get a line #
    ##############

    with open(fname_line, 'r') as fid:
         input_data = json.load(fid)
    line0_no_sc = xt.Line.from_dict(input_data['line'])
    line0_no_sc.particle_ref=xp.Particles.from_dict(input_data['particle'])

    lprofile = xf.LongitudinalProfileQGaussian(
        number_of_particles=bunch_intensity,
        sigma_z=sigma_z,
        z0=0.,
        q_parameter=1.)

    ##################
    # Make particles #
    ##################
    line_temp = line0_no_sc.filter_elements(
        exclude_types_starting_with='SpaceCh')
    line_temp.build_tracker()
    import warnings
    warnings.filterwarnings('ignore')
    particle_probe = line_temp.build_particles(
                weight=0,  # pure probe particles
                zeta=0, delta=0,
                x_norm=2, px_norm=0,
                y_norm=2, py_norm=0,
                nemitt_x=nemitt_x, nemitt_y=nemitt_y)

    particles_gaussian = xp.generate_matched_gaussian_bunch(line=line_temp,
             num_particles= 2 * n_part, # will mark half of them as lost
             total_intensity_particles = 2 * bunch_intensity, # will mark half of them as lost
             nemitt_x=nemitt_x, nemitt_y=nemitt_y, sigma_z=sigma_z)

    particles_gaussian.state[1::2] = -222 # Mark half of them as lost

    particles0 = xp.Particles.merge([particle_probe, particles_gaussian])

    if isinstance(particles0._context, xo.ContextCpu):
        particles0.reorganize()

    warnings.filterwarnings('default')

    if (isinstance(test_context, xo.ContextPyopencl)
            and mode.startswith('pic')):
        # TODO With pyopencl the test gets to the end
        # but then hangs or crashes python
        pytest.skip('This test in broken on OpenCL. Known issue...')
        return

    # We need only particles at zeta close to the probe
    if mode == 'frozen':
        particles = particles0.filter(particles0.particle_id < 100)
    elif mode == 'quasi-frozen':
        particles = particles0.filter(particles0.particle_id < 1e5)
        particles.weight[:] = bunch_intensity / 1e5 * 2 # need to have the right peak density
    elif mode == 'pic' or mode == 'pic_average_transverse':
        particles = particles0.filter((particles0.zeta>z_range[0]*5)
                                      & (particles0.zeta<z_range[1]*5))
    else:
        raise ValueError('Invalid mode!')

    particles.move(_context=test_context)

    warnings.filterwarnings('ignore')
    line = line0_no_sc.copy()
    xf.install_spacecharge_frozen(
            line=line,
            longitudinal_profile=lprofile,
            nemitt_x=nemitt_x, nemitt_y=nemitt_y,
            sigma_z=sigma_z,
            num_spacecharge_interactions=num_spacecharge_interactions)
    warnings.filterwarnings('default')

    # Move to the right context
    line.build_tracker(_context=test_context)
    assert line._context is test_context
    buffer_for_check = line._buffer

    ##########################
    # Configure space-charge #
    ##########################

    if mode == 'frozen':
        pass # Already configured in line
    elif mode == 'quasi-frozen':
        xf.replace_spacecharge_with_quasi_frozen(
                                        line,
                                        update_mean_x_on_track=True,
                                        update_mean_y_on_track=True)
    elif mode == 'pic' or mode == 'pic_average_transverse':
        if mode == 'pic':
            solver = 'FFTSolver2p5D'
        elif mode == 'pic_average_transverse':
            solver = 'FFTSolver2p5DAveraged'
        pic_collection, all_pics = xf.replace_spacecharge_with_PIC(
            line=line,
            n_sigmas_range_pic_x=5,
            n_sigmas_range_pic_y=5,
            nx_grid=256, ny_grid=256, nz_grid=nz_grid,
            n_lims_x=7, n_lims_y=3,
            z_range=z_range,
            solver=solver)
    else:
        raise ValueError(f'Invalid mode: {mode}')

    # rebuild the tracker after editing
    line.build_tracker(_context=test_context)
    assert line._buffer is buffer_for_check

    if mode == 'pic_average_transverse' or mode == 'pic':
        assert isinstance(line[0], xf.SpaceCharge3D)
    else:
        assert isinstance(line[0], xf.SpaceChargeBiGaussian)

    if mode != 'frozen':
        assert line.iscollective
    else:
        assert not line.iscollective

    ###############################
    # Tune shift from single turn #
    ###############################

    line_no_sc = line.filter_elements(exclude_types_starting_with='SpaceCh')
    tw = line_no_sc.twiss(at_elements=[0])

    p_probe_before = particles.filter(
            particles.particle_id == 0).to_dict()

    assert line._buffer is buffer_for_check

    print('Start tracking...')
    line.track(particles)
    print('Done tracking.')

    p_probe_after = particles.filter(
            particles.particle_id == 0).to_dict()

    betx = tw['betx'][0]
    alfx = tw['alfx'][0]
    print(f'{alfx=} {betx=}')
    phasex_0 = np.angle(p_probe_before['x'] / np.sqrt(betx) -
                        1j*(p_probe_before['x'] * alfx / np.sqrt(betx) +
                        p_probe_before['px'] * np.sqrt(betx)))[0]
    phasex_1 = np.angle(p_probe_after['x'] / np.sqrt(betx) -
                       1j*(p_probe_after['x'] * alfx / np.sqrt(betx) +
                           p_probe_after['px'] * np.sqrt(betx)))[0]
    bety = tw['bety'][0]
    alfy = tw['alfy'][0]
    print(f'{alfy=} {bety=}')
    phasey_0 = np.angle(p_probe_before['y'] / np.sqrt(bety) -
                        1j*(p_probe_before['y'] * alfy / np.sqrt(bety) +
                        p_probe_before['py'] * np.sqrt(bety)))[0]
    phasey_1 = np.angle(p_probe_after['y'] / np.sqrt(bety) -
                        1j*(p_probe_after['y'] * alfy / np.sqrt(bety) +
                        p_probe_after['py'] * np.sqrt(bety)))[0]
    qx_probe = (phasex_1 - phasex_0)/(2*np.pi)
    qy_probe = (phasey_1 - phasey_0)/(2*np.pi)

    qx_target = 0.12424673159186882
    qy_target = 0.21993469870358598

    print(f'ex={(qx_probe - qx_target)/1e-3:.6f}e-3 '
          f'ey={(qy_probe - qy_target)/1e-3:.6f}e-3')

    xo.assert_allclose(qx_probe, qx_target, atol=5e-4, rtol=0)
    xo.assert_allclose(qy_probe, qy_target, atol=5e-4, rtol=0)

    if mode == 'pic_average_transverse':
        sc_test = all_pics[50]
        assert sc_test.fieldmap.solver.__class__.__name__ == 'FFTSolver2p5DAveraged'
        ctx2np = sc_test._context.nparray_from_context_array
        for dtest in [sc_test.fieldmap.dphi_dx, sc_test.fieldmap.dphi_dy]:
            # Check that the normalized electric field is the same
            dtest = ctx2np(dtest)
            xo.assert_allclose(
                dtest[:, :, 3] / np.max(dtest[:, :, 3]),
                dtest[:, :, 4] / np.max(dtest[:, :, 4]),
                atol=1e-10, rtol=1e-5)
    elif mode == 'pic':
        sc_test = all_pics[50]
        assert sc_test.fieldmap.solver.__class__.__name__ == 'FFTSolver2p5D'
